import logging
import os
import sys
import time
from logging.handlers import RotatingFileHandler, TimedRotatingFileHandler
from unittest.mock import patch

import pytest

import runez
from runez.ascii import AsciiAnimation, AsciiFrames
from runez.conftest import WrappedHandler
from runez.logsetup import _formatted_text, formatted, LogSpec


LOG = logging.getLogger(__name__)


def test_allow_root(temp_log, monkeypatch):
    with patch("runez.logsetup.os.geteuid", return_value=0):
        # Default: don't complain about running as root inside docker
        monkeypatch.setattr(runez.SYS_INFO, "is_running_in_docker", True)
        runez.log.setup()
        assert "should not be ran as root" not in temp_log.stderr.pop()

        # Complain if explicitly disallowed
        runez.log.setup(allow_root=False)
        assert "should not be ran as root" in temp_log.stderr.pop()

        # Default: complain about running as root outside of docker
        monkeypatch.setattr(runez.SYS_INFO, "is_running_in_docker", False)
        runez.log.setup()
        assert "====\npytest should not be ran as root" in temp_log.stderr.pop()

        # Abort execution with 'allow_root=None'
        with pytest.raises(runez.system.AbortException):
            runez.log.setup(allow_root=None)
        assert "should not be ran as root" in temp_log.stderr.pop()

        # If message doesn't end with '!', then don't do the multi-line '====...' bar decoration
        prev = runez.log.disallow_root_message
        runez.log.disallow_root_message = "no-root-plz"
        runez.log.setup()
        assert "====" not in temp_log.stderr
        assert "no-root-plz" in temp_log.stderr.pop()
        runez.log.disallow_root_message = prev


def test_auto_location_not_writable(temp_log):
    with patch("runez.file.os.access", return_value=False):
        runez.log.setup(
            greetings="Logging to: {location}",
            console_format="%(name)s f:%(filename)s mod:%(module)s func:%(funcName)s %(levelname)s - %(message)s",
            console_level=logging.DEBUG,
        )
        logging.info("hello")
        assert "runez f:logsetup.py mod:logsetup func:greet DEBUG" in temp_log.stderr
        assert "INFO - hello" in temp_log.stderr
        assert "Logging to: no usable locations" in temp_log.stderr
        assert runez.log.file_handler is None


def test_clean_handlers(temp_log):
    # Initially, only pytest logger is here
    assert WrappedHandler.count_non_wrapped_handlers() == 0
    existing = len(logging.root.handlers)

    # Default setup adds a console + file log
    runez.log.setup()
    assert WrappedHandler.count_non_wrapped_handlers() == 2
    assert len(logging.root.handlers) == existing + 2

    # Clean up all non-runez handlers: removes pytest's handler
    runez.log.setup(clean_handlers=True)
    assert WrappedHandler.count_non_wrapped_handlers() == 2
    assert len(logging.root.handlers) == 2

    # Cancel file log
    runez.log.setup(file_format=None)
    assert WrappedHandler.count_non_wrapped_handlers() == 1
    assert len(logging.root.handlers) == 1


def test_console(temp_log):
    logger = logging.getLogger("runez")
    old_level = logger.level

    try:
        runez.log.setup(console_level=logging.DEBUG, file_location="", greetings="Logging to: {location}, argv: {argv}")

        assert temp_log.logfile is None
        assert "DEBUG Logging to: file log disabled" in temp_log.stderr
        assert ", argv: " in temp_log.stderr
        logger.info("hello")
        assert "INFO hello" in temp_log.stderr

        temp_log.clear()
        runez.log.silence(runez)
        logger.info("hello")
        assert not temp_log

    finally:
        logger.setLevel(old_level)


def test_context(temp_log):
    runez.log.spec.locations = None
    runez.log.spec.console_stream = sys.stdout
    runez.log.spec.console_format = "%(timezone)s %(context)s%(levelname)s - %(message)s"
    runez.log.spec.console_level = logging.DEBUG
    assert runez.log.spec.dev is None
    assert runez.log.spec.project is None
    runez.log.setup(greetings=None)
    assert runez.log.spec.dev
    assert runez.log.spec.project

    assert temp_log.logfile is None

    # Edge case: verify adding/removing ends up with empty context
    runez.log.context.add_global(x="y")
    runez.log.context.remove_global("x")
    assert not runez.log.context.has_global()

    runez.log.context.add_threadlocal(x="y")
    runez.log.context.remove_threadlocal("x")
    assert not runez.log.context.has_threadlocal()

    # Add a couple global/thread context values
    runez.log.context.set_global(version="1.0", name="foo")
    runez.log.context.add_threadlocal(worker="susan", a="b")
    logging.info("hello")
    assert temp_log.pop() == "UTC [[a=b,name=foo,version=1.0,worker=susan]] INFO - hello"
    assert runez.log.context.has_threadlocal()

    # Remove them one by one
    runez.log.context.remove_threadlocal("a")
    assert runez.log.context.has_threadlocal()
    logging.info("hello")
    assert temp_log.pop() == "UTC [[name=foo,version=1.0,worker=susan]] INFO - hello"

    runez.log.context.remove_global("name")
    logging.info("hello")
    assert temp_log.pop() == "UTC [[version=1.0,worker=susan]] INFO - hello"

    runez.log.context.remove_threadlocal("worker")
    assert not runez.log.context.has_threadlocal()

    runez.log.context.clear_threadlocal()
    logging.info("hello")
    assert temp_log.pop() == "UTC [[version=1.0]] INFO - hello"

    runez.log.context.clear_global()
    logging.info("hello")
    assert temp_log.pop() == "UTC INFO - hello"

    assert not runez.log.context.has_global()
    assert not runez.log.context.has_threadlocal()


def test_convenience(temp_log):
    fmt = "f:%(filename)s mod:%(module)s func:%(funcName)s %(levelname)s %(message)s "
    fmt += " path:%(pathname)s"
    runez.log.setup(console_format=fmt, console_level=logging.DEBUG, file_format=None)

    assert temp_log.logfile is None
    runez.write("some-file", "some content", logger=logging.info)
    logging.info("hello")
    logging.exception("oops")

    assert "f:system.py mod:system func:hlog INFO Wrote some-file" in temp_log.stderr
    assert "f:test_logsetup.py mod:test_logsetup func:test_convenience INFO hello" in temp_log.stderr
    assert "f:test_logsetup.py mod:test_logsetup func:test_convenience ERROR oops" in temp_log.stderr
    temp_log.stderr.clear()

    runez.write("some-file", "some content", logger=LOG.info)
    LOG.info("hello")
    LOG.exception("oops")
    assert "f:system.py mod:system func:hlog INFO Wrote some-file" in temp_log.stderr
    assert "f:test_logsetup.py mod:test_logsetup func:test_convenience INFO hello" in temp_log.stderr
    assert "f:test_logsetup.py mod:test_logsetup func:test_convenience ERROR oops" in temp_log.stderr


def test_default(temp_log):
    assert runez.log.resolved_dryrun(True) is True
    assert runez.log.resolved_dryrun(False) is False
    assert runez.log.resolved_dryrun(runez.UNSET) is False

    assert not runez.log.hdry("do something")
    assert not temp_log.pop()

    assert runez.log.hdry("do something", dryrun=True)
    assert temp_log.pop() == "Would do something"

    assert runez.log.spec.console_level == logging.WARNING
    runez.log.context.set_global(version="1.0")
    runez.log.context.add_global(worker="mark")
    runez.log.context.add_threadlocal(worker="joe", foo="bar")
    runez.log.context.set_threadlocal(worker="joe")
    runez.log.setup(greetings="Logging to: {location}, pid {pid}")

    assert os.path.basename(temp_log.logfile) == "pytest.log"
    temp_log.expect_logged("Logging to: ")
    temp_log.expect_logged("pytest.log, pid %s" % os.getpid())

    logging.info("hello")
    logging.warning("hello")
    temp_log.expect_logged("INFO hello")
    temp_log.expect_logged("WARNING hello")
    assert "INFO hello" not in temp_log.stderr
    assert "WARNING hello" in temp_log.stderr.pop()

    # Now stop logging context
    runez.log.setup(
        console_format="%(funcName)s %(module)s %(levelname)s %(message)s",
        file_format="%(asctime)s %(timezone)s %(levelname)s %(message)s",
    )
    logging.info("hello")
    logging.warning("hello")
    temp_log.expect_logged("UTC INFO hello")
    temp_log.expect_logged("UTC WARNING hello")
    assert temp_log.pop() == "test_default test_logsetup WARNING hello"


def test_deprecated():
    # Test coverage for deprecated functions
    assert runez.log.tests_path() == runez.DEV.tests_path()  # deprecated


@pytest.mark.skipif(runez.SYS_INFO.platform_id.is_windows, reason="No /dev/null on Windows")
def test_file_location_not_writable(temp_log):
    runez.log.setup(
        greetings="Logging to: {location}",
        console_level=logging.DEBUG,
        file_location="/dev/null/somewhere.log",
    )
    assert "DEBUG Logging to: given location '/dev/null/somewhere.log' is not usable" in temp_log.stderr
    assert runez.log.file_handler is None


def test_formatted():
    assert formatted("foo") == "foo"
    assert formatted("foo", "bar") == "foo"  # Ignoring extra positionals

    assert formatted("foo %s", "bar") == "foo bar"
    assert formatted("foo {0}", "bar") == "foo bar"
    assert formatted("foo %s {0}", "bar") == "foo bar {0}"  # '%s' format used first

    assert formatted("foo %s {a}", "bar", a="val_a") == "foo %s val_a"  # '%s' does not apply when there are kwargs

    # Bogus formats
    assert formatted("foo %s %s {0}", "bar") == "foo %s %s bar"  # bogus '%s' format
    assert formatted("foo %s %s {0} {1}", "bar") == "foo %s %s {0} {1}"  # bogus '%s' and {positional} format


def test_formatted_text():
    # Unsupported formats
    assert _formatted_text("", {}) == ""
    assert _formatted_text("{filename}", {}) == "{filename}"
    assert _formatted_text("{filename}", {}, strict=True) is None  # In strict mode, all named refs must be defined
    assert _formatted_text("{filename}", {"filename": "foo"}) == "foo"
    assert _formatted_text("{filename}", {"filename": "foo"}, strict=True) == "foo"
    assert _formatted_text("{foo}/{foo}", {"foo": "bar", "unused": "ok"}) == "bar/bar"
    assert _formatted_text("~/.cache/{foo}", {"foo": "bar"}) == os.path.expanduser("~/.cache/bar")

    # Verify that not all '{...}' text is considered a marker
    props = dict(argv='{a} {"foo": "bar {a}"} {a}', a="{b}", b="b")
    assert _formatted_text(":: {argv} {a}", props) == ':: {a} {"foo": "bar {a}"} {a} b'

    deep = dict(a="a", b="b", aa="{a}", bb="{b}", ab="{aa}{bb}", ba="{bb}{aa}", abba="{ab}{ba}", deep="{abba}")
    assert _formatted_text("{deep}", deep, max_depth=-1) == "{deep}"
    assert _formatted_text("{deep}", deep, max_depth=0) == "{deep}"
    assert _formatted_text("{deep}", deep, max_depth=1) == "{abba}"
    assert _formatted_text("{deep}", deep, max_depth=2) == "{ab}{ba}"
    assert _formatted_text("{deep}", deep, max_depth=3) == "{aa}{bb}{bb}{aa}"
    assert _formatted_text("{deep}", deep, max_depth=4) == "{a}{b}{b}{a}"
    assert _formatted_text("{deep}", deep, max_depth=5) == "abba"
    assert _formatted_text("{deep}", deep, max_depth=6) == "abba"

    recursive = dict(a="a{b}", b="b{c}", c="c{a}")
    assert _formatted_text("{a}", recursive) == "abc{a}"
    assert _formatted_text("{a}", recursive, max_depth=10) == "abcabcabca{b}"

    cycle = dict(a="{b}", b="{a}")
    assert _formatted_text("{a}", cycle, max_depth=0) == "{a}"
    assert _formatted_text("{a}", cycle, max_depth=1) == "{b}"
    assert _formatted_text("{a}", cycle, max_depth=2) == "{a}"
    assert _formatted_text("{a}", cycle, max_depth=3) == "{b}"


def test_level(temp_log):
    runez.log.setup(file_format=None, level=logging.INFO)

    assert not temp_log
    assert temp_log.logfile is None
    logging.debug("debug msg")
    logging.info("info msg")
    assert "debug msg" not in temp_log.stderr
    assert "info msg" in temp_log.stderr


def test_locations(temp_log):
    runez.log.setup(locations=None)
    assert runez.log.file_handler is None

    # Verify that a non-writeable folder is not used
    runez.ensure_folder("foo/bar", logger=None)
    os.chmod("foo/bar", 0o400)
    runez.log.setup(locations=["foo/bar"])
    os.chmod("foo/bar", 0o700)
    assert runez.log.file_handler is None

    runez.log.setup(locations=["{dev}/test-location.log"])
    assert runez.log.file_handler

    runez.log.setup(locations=["{project}/.venv/test-location.log"])
    assert runez.log.file_handler


def test_log_rotate(temp_folder):
    with pytest.raises(ValueError):
        runez.log.setup(rotate="foo", tmp=temp_folder)

    with pytest.raises(ValueError):
        runez.logsetup._get_file_handler("test.log", "time", 0)

    with pytest.raises(ValueError):
        runez.logsetup._get_file_handler("test.log", "time:unclear", 0)

    with pytest.raises(ValueError):
        runez.logsetup._get_file_handler("test.log", "time:h", 0)

    with pytest.raises(ValueError):
        runez.logsetup._get_file_handler("test.log", "time:1h,something", 0)

    with pytest.raises(ValueError):
        runez.logsetup._get_file_handler("test.log", "size:not a number,3", 0)

    with pytest.raises(ValueError):
        runez.logsetup._get_file_handler("test.log", "unknown:something", 0)

    assert runez.logsetup._get_file_handler("test.log", None, 0).__class__ is logging.FileHandler
    assert runez.logsetup._get_file_handler("test.log", "", 0).__class__ is logging.FileHandler

    h = runez.logsetup._get_file_handler("test.log", "time:1h", 0)
    assert isinstance(h, TimedRotatingFileHandler)
    assert h.backupCount == 0
    assert h.interval == 3600
    assert h.when == "H"

    h = runez.logsetup._get_file_handler("test.log", "time:midnight", 7)
    assert isinstance(h, TimedRotatingFileHandler)
    assert h.backupCount == 7
    assert h.when == "MIDNIGHT"

    h = runez.logsetup._get_file_handler("test.log", "size:10k", 3)
    assert isinstance(h, RotatingFileHandler)
    assert h.backupCount == 3
    assert h.maxBytes == 10240


def test_logspec(isolated_log_setup):
    s1 = LogSpec(runez.log._default_spec, appname="pytest")
    s2 = LogSpec(runez.log._default_spec, appname="pytest")
    assert s1 == s2
    assert s1.appname == "pytest"
    assert s1.timezone == "UTC"
    assert s1.should_log_to_file

    # No basename -> can't determine a usable location anymore
    s1.basename = None
    assert s1.should_log_to_file
    assert s1.usable_location() is None

    s1.set(basename="testing.log", timezone=None, locations=[s1.tmp])
    assert s1.basename == "testing.log"
    assert s1.timezone is None
    assert s1 != s2

    # Empty string custom location just disables file logging
    s1.file_location = ""
    assert not s1.should_log_to_file
    assert s1.usable_location() is None

    # No basename, and custom location points to folder -> not usable
    s1.basename = None
    s1.file_location = "./foo"
    assert s1.should_log_to_file
    assert s1.usable_location() == "./foo"

    # Restore from other spec
    s1.set(s2)
    assert s1 == s2

    s1.set(s2, timezone="hello")
    assert s1 != s2
    assert s1.timezone == "hello"

    s1.set(s2, timezone=runez.UNSET)
    assert s1 == s2

    # No-ops, because targets don't have any meaningful values
    s1.set(not_valid="this is not a field of LogSpec")
    assert s1 == s2

    s1.set("hello")
    assert s1 == s2

    s1.set(s2, "hello")
    assert s1 == s2


def test_no_context(temp_log):
    runez.log.context.set_global(version="1.0")
    runez.log.spec.set(timezone="", file_format="%(asctime)s [%(threadName)s] %(timezone)s %(levelname)s - %(message)s")
    runez.log.setup()
    logging.info("hello")
    temp_log.expect_logged("[MainThread] INFO - hello")


def test_setup(temp_log, monkeypatch):
    fmt = "%(asctime)s %(context)s%(levelname)s - %(message)s"
    assert runez.log.is_using_format("", fmt) is False
    assert runez.log.is_using_format("%(lineno)", fmt) is False
    assert runez.log.is_using_format("%(context)", fmt) is True
    assert runez.log.is_using_format("%(context) %(lineno)", fmt) is True
    assert runez.log.is_using_format("%(context)", "") is False

    if not runez.SYS_INFO.platform_id.is_windows:
        # signum=None is equivalent to disabling faulthandler
        runez.log.enable_faulthandler(signum=None)
        assert runez.log.faulthandler_signum is None
        # We didn't call setup, so enabling faulthandler will do nothing
        runez.log.enable_faulthandler()
        assert runez.log.faulthandler_signum is None

    cwd = os.getcwd()
    assert not runez.DRYRUN
    assert not runez.log.debug
    with runez.TempFolder(dryrun=False):
        assert not runez.log.debug

        # No auto-debug on dryrun
        runez.log.setup(dryrun=True, level=logging.INFO)
        runez.log.enable_trace(None)
        assert not runez.log.debug
        assert runez.DRYRUN
        logging.info("info")
        logging.debug("hello")
        runez.log.trace("some trace info")
        assert not temp_log.stdout
        assert "hello" not in temp_log
        assert "some trace info" not in temp_log  # Tracing not enabled
        assert "info" in temp_log.stderr.pop()

        # Second call without any customization is a no-op
        runez.log.setup()
        runez.log.enable_trace(False)
        assert not runez.log.debug
        assert runez.DRYRUN
        logging.debug("hello")
        runez.log.trace("some trace info")
        assert "some trace info" not in temp_log  # Tracing not enabled
        assert not temp_log

        # Change stream
        runez.log.setup(console_stream=sys.stdout, trace="SOME_ENV_VAR")
        logging.info("hello")
        runez.log.trace("some trace info")
        assert not temp_log.stderr
        assert "some trace info" not in temp_log  # Not tracing because env var not set
        assert "INFO hello" in temp_log.stdout.pop()

        # Change logging level
        runez.log.setup(console_level=logging.WARNING, trace=True)
        logging.info("hello")
        assert not temp_log
        logging.warning("hello")
        runez.log.trace("some trace %s", "info")
        assert ":: some trace info" in temp_log  # Tracing forcibly enabled
        assert "WARNING hello" in temp_log.stdout.pop()
        assert not temp_log.stderr

        # Change format and enable debug + tracing
        monkeypatch.setenv("SOME_ENV_VAR", "1")
        runez.log.setup(debug=True, console_format="%(levelname)s - %(message)s", trace="SOME_ENV_VAR+... ")
        assert runez.log.debug
        assert runez.log.console_handler.level == logging.DEBUG
        logging.debug("hello")
        runez.log.trace("some trace info")
        assert "... some trace info" in temp_log  # We're now tracing (because env var is set)
        assert "DEBUG - hello" in temp_log.stdout.pop()
        assert not temp_log.stderr

        if not runez.SYS_INFO.platform_id.is_windows and runez.logsetup.faulthandler:
            # Available only in python3
            runez.log.enable_faulthandler()
            assert runez.log.faulthandler_signum

        assert runez.log.debug is True
        assert runez.DRYRUN is True

    # Verify dryrun and current folder restored, but debug untouched
    assert runez.log.debug
    assert not runez.DRYRUN
    assert os.getcwd() == cwd


def test_progress_bar():
    p = runez.ProgressBar(range(2))
    assert list(p) == [0, 1]
    assert str(p) == "None/2"
    assert p.rendered() is None

    with runez.ProgressBar(total=3, columns=4) as pb:
        assert pb.n == 0
        assert pb.rendered() == "    0%"
        pb.update()
        assert pb.n == 1
        assert pb.rendered() == "▉▏  33%"
        pb.update()
        assert pb.rendered() == "▉▉▌ 67%"
        pb.update()
        assert pb.rendered() == "▉▉▉▉100%"

    assert pb.n is None
    assert pb.rendered() is None


def test_progress_command(cli, monkeypatch):
    cli.run("progress-bar", "-i10", "-d1", "--sleep", "0.01")
    assert cli.succeeded
    assert "done" in cli.logged.stdout
    assert "CPU usage" in cli.logged.stdout

    monkeypatch.setitem(sys.modules, "psutil", None)
    cli.run("progress-bar", "-i10", "-d1", "--sleep", "0.01")
    assert cli.succeeded
    assert "done" in cli.logged.stdout
    assert "CPU usage" not in cli.logged


def test_progress_frames(monkeypatch):
    foo = AsciiFrames(["a", ["b", ""]], fps=10)
    assert foo.frames == ["a", "b"]
    assert foo.index == 0
    assert foo.next_frame() == "b"
    assert foo.index == 1
    assert foo.next_frame() == "a"
    assert foo.index == 0
    assert foo.next_frame() == "b"

    assert AsciiAnimation.get_frames(None)
    assert AsciiAnimation.get_frames(AsciiAnimation.af_dots)
    f = AsciiFrames(list("ab"))
    assert AsciiAnimation.get_frames(f) is f
    assert AsciiAnimation.from_spec(f) is f

    assert AsciiAnimation.predefined("foo") is None
    assert AsciiAnimation.predefined("random")
    off = AsciiAnimation.predefined("off")
    assert off.frames is None
    assert str(off) == "off"
    names = AsciiAnimation.available_names()
    assert names
    for name in names:
        assert AsciiAnimation.predefined(name)

    monkeypatch.setattr(AsciiAnimation, "env_var", None)
    monkeypatch.setattr(AsciiAnimation, "default", None)
    f = AsciiAnimation.get_frames(None)
    assert not f.frames


def next_progress_line(progress_spinner):
    ts = getattr(progress_spinner, "_test_ts", None)
    if ts is None:
        ts = time.time()

    ts += 1
    text = progress_spinner._state.get_line(ts)
    setattr(progress_spinner, "_test_ts", ts)
    return text


def test_progress_grooming():
    # Verify that short messages with newlines get their newline changed to a space
    p = runez.logsetup.ProgressSpinner()
    p._state = runez.logsetup._SpinnerState(p, AsciiFrames(None), 80, None, None, None)
    assert next_progress_line(p) is None
    p.show("\n a \n\n\n b \n")
    assert next_progress_line(p) == " a b"


def test_progress_operation(isolated_log_setup, logged):
    assert not runez.log.progress.is_running
    runez.log.progress.start()
    assert not runez.log.progress.is_running  # Does not start in test mode by default

    runez.log.progress.stop()  # no-op, already not running
    assert not runez.log.progress.is_running

    runez.log.setup()
    logged.clear()
    with patch("runez.system.TerminalInfo.isatty", return_value=True):
        # Simulate progress with alternating foo/bar "spinner", using `str` to cover color code path
        p1 = runez.ProgressBar(total=3)
        p1.start()
        p1.update()
        assert runez.log.progress._progress_bar
        runez.log.progress.start(max_columns=10, message_color=str, spinner_color=str)
        assert runez.log.progress.is_running
        assert runez.log.progress._progress_bar
        logging.error("some error")
        time.sleep(0.1)
        runez.log.progress.show("some progress")
        assert runez.log.progress._progress_bar
        print("hello")
        time.sleep(0.1)
        p2 = runez.ProgressBar(total=3)
        p3 = runez.ProgressBar(total=3)
        p4 = runez.ProgressBar(total=3)
        p1.start()
        p2.start()
        p3.start()
        p4.start()

        assert runez.log.progress._progress_bar is p4
        time.sleep(0.1)

        # Stop progress bar in a different order
        p2.stop()
        assert runez.log.progress._progress_bar is p4
        p4.stop()
        assert runez.log.progress._progress_bar is p3
        p3.stop()
        assert runez.log.progress._progress_bar is p1
        p1.stop()
        assert runez.log.progress._progress_bar is None

        runez.log.progress.stop()
        assert not runez.log.progress.is_running

        assert "hello" in logged.stdout
        assert "some error" in logged.stderr

        # Simulate progress without spinner
        logged.clear()
        runez.log.progress.start(frames=None)
        runez.log.progress.show("some progress")
        time.sleep(0.1)
        assert runez.log.progress.is_running
        runez.log.progress.stop()
        time.sleep(0.1)
        assert not runez.log.progress.is_running


class SampleClass:

    @runez.log.timeit  # Without args
    def instance_func1(self, message, fail=False):
        if fail:
            raise ValueError("oops")

        print("%s: %s" % (self, message))

    @runez.log.timeit()  # With args
    def instance_func2(self, message):
        print("%s: %s" % (self, message))

    @classmethod
    @runez.log.timeit
    def class_func1(cls, message):
        print("%s: %s" % (cls, message))

    @classmethod
    @runez.log.timeit()
    def class_func2(cls, message):
        print("%s: %s" % (cls, message))

    @staticmethod
    @runez.log.timeit()
    def static_func1(message):
        print(message)

    @staticmethod
    @runez.log.timeit()
    def static_func2(message):
        print(message)


@runez.log.timeit
def sample_function1(message):
    print(message)


@runez.log.timeit("sample2", color=True, logger=print)
def sample_function2(message):
    print(message)


def test_timeit(logged):
    sample = SampleClass()
    sample.instance_func1("hello")
    assert "SampleClass.instance_func1() took " in logged.pop()

    with pytest.raises(ValueError):
        sample.instance_func1("hello", fail=True)
    assert "SampleClass.instance_func1() failed: oops" in logged.pop()

    sample.instance_func2("hello")
    assert "SampleClass.instance_func2() took " in logged.pop()

    sample.class_func1("hello")
    assert "SampleClass.class_func1() took " in logged.pop()

    sample.class_func2("hello")
    assert "SampleClass.class_func2() took " in logged.pop()

    sample.static_func1("hello")
    assert "SampleClass.static_func1() took " in logged.pop()

    sample.static_func2("hello")
    assert "SampleClass.static_func2() took " in logged.pop()

    sample_function1("sample1")
    assert "sample_function1() took " in logged.pop()

    sample_function2("sample2")
    assert "sample2 took " in logged.pop()

    with runez.log.timeit():
        print("hello")
    assert "tests.test_logsetup.test_timeit() took " in logged.pop()

    with runez.log.timeit("ad-hoc context"):
        print("hello")
    assert "ad-hoc context took " in logged.pop()



import os
import sys
from argparse import Namespace
from unittest.mock import patch

import pytest

from runez.render import Align, Header, PrettyBorder, PrettyHeader, PrettyTable
from runez.system import SYS_INFO, SystemInfo, UNSET


def test_align():
    assert Align.cast("left") is Align.left
    assert Align.cast("CENTER") is Align.center
    assert Align.cast("foo", default=None) is None
    assert Align.cast("foo", default=Align.center) is Align.center
    assert Align.cast("foo", default="center") is Align.center
    assert Align.cast(None, default=Align.right) is Align.right
    assert Align.cast(Align.right) is Align.right

    with pytest.raises(ValueError):
        assert Align.cast("foo")

    with pytest.raises(ValueError):
        assert Align.cast("foo", "foo")

    assert Align.center("foo", 5) == " foo "
    assert Align.left("foo", 5) == "foo  "
    assert Align.left("foo", 5, fill="-") == "foo--"
    assert Align.right("foo", 5) == "  foo"


def test_border():
    t1 = PrettyBorder("empty")
    t2 = PrettyBorder()
    assert t1 == t2
    assert str(t1) == "c:,pad:1"

    assert PrettyBorder.cast(t1) is t1
    assert PrettyBorder.cast("empty") == t1

    tc = PrettyBorder("compact")
    assert str(tc) == "c:   ,h:   -,pad:1"

    # Exercise setting from object fields, for coverage
    tc2 = PrettyBorder("c:   ", h=Namespace(first=" ", mid=" ", last=" ", h="-"))
    assert tc2 == tc


EXPECTED_DIAGNOSTICS = """
  foo : bar
  opt : -missing-
  --- :

Other section:
    --- :
  diag2 : UNSET

Some section:
   key1 : value1
   key2 : -missing-
    --- :
  diag2 : UNSET

some additional text
"""


def test_diagnostics(monkeypatch):
    def _diag1(opt=None):
        yield "foo", "bar"
        yield "opt", opt
        yield "---"

    def _diag2():
        yield "---"
        yield "diag2", UNSET

    data = {"key1": "value1", "key2": None}
    sections = {"Some section": [data, _diag2], "Other section": _diag2}
    diag1 = PrettyTable.two_column_diagnostics(_diag1, sections, "some additional text")
    assert diag1 == EXPECTED_DIAGNOSTICS.strip("\n")

    # Same, but with calling the generators first
    sections = {"Some section": [data, _diag2()], "Other section": _diag2()}
    diag1 = PrettyTable.two_column_diagnostics(_diag1(), sections, "some additional text")
    assert diag1 == EXPECTED_DIAGNOSTICS.strip("\n")

    diag2 = PrettyTable.two_column_diagnostics(_diag2(), SYS_INFO.diagnostics())
    assert "diag2 : UNSET" in diag2
    assert "sys.executable" in diag2
    assert "invoker" in diag2  # Present when running from venv

    silly_edge_case = PrettyTable.two_column_diagnostics(0, 1, [2])
    assert silly_edge_case == "  0 :\n  1 :\n  2 :"

    with patch.dict(os.environ, {"LC_TERMINAL": "foo", "LC_TERMINAL_VERSION": "2"}):
        monkeypatch.setattr(sys, "executable", "foo")
        s = SystemInfo()
        x = PrettyTable.two_column_diagnostics(s.diagnostics())
        assert "terminal : foo v2" in x
        assert "via : " in x

        x = PrettyTable.two_column_diagnostics(s.diagnostics(via=None))
        assert "via : " not in x


def test_header():
    assert Header.aerated("") == ""
    assert Header.aerated("test", border="") == "test"

    assert Header.aerated("test", border="--") == "----------\n-- test --\n----------"
    assert Header.aerated("test", border="-- ") == "-- test\n-------"

    assert Header.aerated("test", border="=") == "========\n= test =\n========"
    assert Header.aerated("test", border="= ") == "= test\n======"


def test_pretty_table():
    with pytest.raises(ValueError):
        PrettyTable(object)  # Invalid header

    t = PrettyTable()
    assert len(t.header) == 0
    with pytest.raises(IndexError):
        t.header[0].style = "bold"  # Header columns are not auto-accommodated when accessing by index

    t.header.accommodate(2)
    assert len(t.header) == 2
    t.header[0].style = "bold"

    t = PrettyTable("1,2,,3", border="pad:0")
    assert len(t.header.columns) == 4
    t.add_rows("a b c".split(), "d e foo".split())

    t.header = 3  # Interpreted as 3 columns (but no header text)
    assert t.get_string() == "abc\ndefoo"

    t.header = [1, 2, 3]  # Numbers will be stringified
    assert isinstance(t.header, PrettyHeader)
    t.align = "left"
    assert t.get_string() == "123\nabc\ndefoo"

    t.align = "center"
    assert t.get_string() == "12 3\nab c\ndefoo"

    t.align = "right"
    assert t.get_string() == "12  3\nab  c\ndefoo"

    t.border = "ascii"
    s = t.get_string()
    assert s == "+===+===+=====+\n| 1 | 2 |   3 |\n+===+===+=====+\n| a | b |   c |\n+---+---+-----+\n| d | e | foo |\n+---+---+-----+"

    t.border = "empty"
    t.header.add_columns("x", "y")
    s = t.get_string()
    assert s == " 1  2    3  x  y\n a  b    c  -  -\n d  e  foo  -  -"

    t.missing = "*"
    t.style = "bold"
    assert t.header["x"] is t.header.columns[-2]

    assert t.header["x"].shown is True
    t.header.hide(-2, 0, "y")
    assert t.header[0].shown is False
    assert t.header["x"].shown is False
    assert t.header["y"].shown is False

    s = t.get_string()
    assert s == " 2    3\n b    c\n e  foo"

    assert t.header[0].shown is False
    with pytest.raises(KeyError):
        t.header.show(0, "foo")

    # Successfully shown column 0, but failed after that on "foo"
    assert t.header[0].shown is True

    assert str(t.header[0]) == "[c0] '1'"
    t.header[0].text = None
    assert str(t.header[0]) == "[c0]"

    with pytest.raises(IndexError):
        _ = t.header.columns[-55]

    with pytest.raises(KeyError):
        _ = t.header["foo"]


EXPECTED_TABLE_RENDER = """
+==========+====================+
| Location | French café        |
+----------+--------------------+
| Name     | コンニチハ, セカイ |
+----------+--------------------+
"""


def test_pretty_table_render():
    t = PrettyTable(border="ascii")
    t.add_row("Location", "French café")
    t.add_row("Name", "コンニチハ, セカイ")
    x = t.get_string()
    assert x == EXPECTED_TABLE_RENDER.strip()



import datetime
import time

from freezegun import freeze_time

import runez


def check_date(expected, dt):
    actual = dt.strftime("%Y-%m-%d %H:%M:%S %Z").strip()
    assert actual == expected


def test_date_formats():
    ref = dt(2019, 1, 16)
    assert runez.to_date("2019-01-16") == ref
    assert runez.to_date("  2019/01/16 ") == ref
    assert runez.to_date("01/16/2019") == ref

    assert runez.to_date("2019- 01-16") is None
    assert runez.to_date("16/01/2019") is None
    assert runez.to_datetime("2019/01/2019") is None

    with freeze_time("2019-09-01 20:00:12"):
        assert str(runez.to_date(" 1w")) == "2019-08-25"
        assert str(runez.to_date("1y")) == "2018-09-01"
        assert str(runez.to_date("1y1w")) == "2018-08-25"
        assert str(runez.to_date(" 1y 1w ")) == "2018-08-25"

        assert str(runez.to_datetime("1y", tz=runez.date.UTC)) == "2018-09-01 14:11:00+00:00"


def test_elapsed():
    d1 = datetime.date(2019, 9, 1)
    dt27 = datetime.datetime(2019, 9, 1, second=27)

    assert runez.to_date(d1) is d1
    assert runez.to_datetime(dt27) is dt27

    assert runez.elapsed(d1, ended=dt27) == 27
    assert runez.elapsed(dt27, ended=d1) == -27

    d2 = datetime.date(2019, 9, 2)
    dt1 = datetime.datetime(2019, 9, 1)
    assert runez.elapsed(d1, ended=d2) == 86400
    assert runez.elapsed(d2, ended=dt1) == -86400

    dt = runez.datetime_from_epoch(1567296012, tz=None)  # Naive date will depend on timezone (ie: where this test runs)
    assert dt.year == 2019
    assert dt.tzinfo is None
    assert runez.to_datetime(dt) is dt

    check_date("2019-09-01 02:00:12 +02:00", runez.datetime_from_epoch(1567296012, tz=runez.timezone_from_text("0200", default=None)))
    check_date("2019-09-01 00:00:12 UTC", runez.datetime_from_epoch(1567296012, tz=runez.date.UTC))
    check_date("2019-09-01 00:00:12 UTC", runez.datetime_from_epoch(1567296012000, tz=runez.date.UTC, in_ms=True))

    with freeze_time("2019-09-01 00:00:12"):
        assert runez.elapsed(datetime.datetime(2019, 9, 1, second=34)) == -22
        assert runez.elapsed(datetime.datetime(2019, 9, 1)) == 12


def test_epoch():
    assert runez.to_epoch(None) is None

    d = datetime.date(2019, 9, 1)
    dt27 = datetime.datetime(2019, 9, 1, second=27, tzinfo=runez.date.UTC)
    assert runez.to_epoch(d) == 1567296000
    assert runez.to_epoch(d, in_ms=True) == 1567296000000
    assert runez.to_epoch(dt27) == 1567296027
    assert runez.to_epoch_ms(dt27) == 1567296027000


def test_represented_duration():
    assert runez.represented_duration(None) == "None"
    assert runez.represented_duration("foo") == "foo"  # noqa, verifiy non-duration left as-is...
    assert runez.represented_duration(runez.UNSET) == "UNSET"  # noqa

    assert runez.represented_duration(0) == "0 seconds"
    assert runez.represented_duration(1) == "1 second"
    assert runez.represented_duration(-1.00001) == "1 second 10 μs"
    assert runez.represented_duration(-180.00001) == "3 minutes"
    assert runez.represented_duration(-180.00001, span=None) == "3 minutes 10 μs"
    assert runez.represented_duration(5.1) == "5 seconds 100 ms"
    assert runez.represented_duration(180.1) == "3 minutes"

    assert runez.represented_duration(65) == "1 minute 5 seconds"
    assert runez.represented_duration(65, span=-2) == "1m 5s"
    assert runez.represented_duration(3667, span=-2) == "1h 1m"
    assert runez.represented_duration(3667, span=None) == "1 hour 1 minute 7 seconds"

    h2 = 2 * runez.date.SECONDS_IN_ONE_HOUR
    d8 = 8 * runez.date.SECONDS_IN_ONE_DAY
    a_week_plus = d8 + h2 + 13 + 0.00001
    assert runez.represented_duration(a_week_plus, span=None) == "1 week 1 day 2 hours 13 seconds 10 μs"
    assert runez.represented_duration(a_week_plus, span=-2, delimiter="+") == "1w+1d"
    assert runez.represented_duration(a_week_plus, span=3) == "1 week 1 day 2 hours"
    assert runez.represented_duration(a_week_plus, span=0) == "1w 1d 2h 13s 10μs"

    five_weeks_plus = (5 * 7 + 3) * runez.date.SECONDS_IN_ONE_DAY + runez.date.SECONDS_IN_ONE_HOUR + 5 + 0.0002
    assert runez.represented_duration(five_weeks_plus, span=-2, delimiter=", ") == "5w, 3d"
    assert runez.represented_duration(five_weeks_plus, span=0, delimiter=", ") == "5w, 3d, 1h, 5s, 200μs"

    assert runez.represented_duration(752 * runez.date.SECONDS_IN_ONE_DAY, span=3) == "2 years 3 weeks 1 day"


def test_timezone(monkeypatch):
    assert runez.local_timezone() == time.tzname[0]
    monkeypatch.setattr(runez.date.time, "tzname", [])
    assert runez.local_timezone() == ""

    assert runez.timezone_from_text(None) is runez.date.DEFAULT_TIMEZONE
    assert runez.timezone_from_text("foo", default=None) is None
    assert runez.timezone_from_text("-00: 00", default=None) is None
    assert runez.timezone_from_text(" +00 00", default=None) is None

    assert runez.timezone_from_text(" Z", default=None) == runez.date.UTC
    assert runez.timezone_from_text("UTC ", default=None) == runez.date.UTC
    assert runez.timezone_from_text(" 0000  ", default=None) == runez.date.UTC
    assert runez.timezone_from_text("+0000", default=None) == runez.date.UTC
    assert runez.timezone_from_text(" -00:00", default=None) == runez.date.UTC
    assert runez.timezone_from_text("+0100", default=None) != runez.date.UTC

    epoch = 1568332800
    assert runez.to_date(epoch) == dt(2019, 9, 13)
    assert runez.to_date(epoch) == runez.to_date(epoch * 1000)
    assert runez.to_datetime(epoch) == dt(2019, 9, 13, 0, 0, 0)

    epoch = 1568348000
    check_date("2019-09-13 04:13:20 UTC", runez.datetime_from_epoch(epoch))

    assert runez.to_date(epoch) == dt(2019, 9, 13)
    assert runez.to_datetime(epoch) == dt(2019, 9, 13, 4, 13, 20)

    tz1 = runez.timezone(datetime.timedelta(seconds=12 * 60))
    dtutc = runez.datetime_from_epoch(epoch, tz=runez.date.UTC)
    dt1 = runez.datetime_from_epoch(epoch, tz=tz1)
    eutc = runez.to_epoch(dtutc)
    et1 = runez.to_epoch(dt1)
    assert et1 - eutc == 720
    assert dtutc == dt1

    tz = runez.timezone_from_text("-01:00", default=None)
    assert str(tz) == "-01:00"
    check_date("2019-09-13 03:13:20 -01:00", runez.datetime_from_epoch(epoch, tz=tz))

    tz = runez.timezone_from_text("0200", default=None)
    assert str(tz) == "+02:00"
    check_date("2019-09-13 06:13:20 +02:00", runez.datetime_from_epoch(epoch, tz=tz))


def dt(*args, **kwargs):
    if len(args) == 3:
        return datetime.date(*args)

    kwargs.setdefault("tzinfo", runez.date.DEFAULT_TIMEZONE)
    return datetime.datetime(*args, **kwargs)


def test_to_date():
    assert runez.to_date("") is None
    assert runez.to_date("   ") is None
    assert runez.to_datetime("") is None
    assert runez.to_datetime("  ") is None

    tz1 = runez.timezone(datetime.timedelta(seconds=12 * 60))
    assert runez.to_datetime("2019-01-02 03:04:05").tzinfo is runez.date.DEFAULT_TIMEZONE
    assert runez.to_datetime("2019-01-02 03:04:05", tz=tz1).tzinfo is tz1
    assert runez.to_datetime("2019-01-02 03:04:05 -00:12").tzinfo == tz1
    assert runez.to_datetime("2019-01-02 03:04:05 UTC").tzinfo is runez.date.UTC

    d0 = runez.to_datetime("2019-01-02 03:04:05  UTC")
    d1 = runez.to_datetime(" 2019-01-02 03:04:05 -00:00 ")
    d2 = runez.to_datetime("2019-01-02 04:04:05 +01:00")
    assert d0 == d2
    assert d1 == d2

    assert runez.to_date(d0) == dt(2019, 1, 2)

    assert runez.to_datetime(None) is None
    assert runez.to_datetime("foo") is None
    assert runez.to_datetime(["foo"]) is None

    assert runez.to_date("2019-01-02") == dt(2019, 1, 2)
    assert runez.to_date("2019-01-02 00:01:00UTC") == dt(2019, 1, 2)
    assert runez.to_date("2019-01-02 00:01:00 UTC") == dt(2019, 1, 2)
    assert runez.to_date("2019-01-02 00:01:00  UTC ") == dt(2019, 1, 2)
    assert runez.to_datetime("2019-01-02") == dt(2019, 1, 2, 0, 0, 0)
    assert runez.to_datetime("2019-01-02 00:01:00 UTC") == dt(2019, 1, 2, 0, 1, 0)

    sample_date = dt(2019, 1, 2, 3, 4, 5, microsecond=678900)
    assert runez.to_datetime("2019-01-02T03:04:05.6789") == sample_date
    assert runez.to_datetime("2019-01-02 03:04:05.6789Z") == sample_date
    assert runez.to_datetime("2019-01-02 03:04:05.6789 Z") == sample_date
    assert runez.to_datetime("2019-01-02 03:04:05.6789 UTC") == sample_date
    assert runez.to_datetime("2019-01-02 03:04:05.6789 -00:00") == sample_date
    assert runez.to_datetime("2019-01-02 04:04:05.6789 +01:00") == sample_date

    assert runez.to_datetime("1500620000") == dt(2017, 7, 21, 6, 53, 20)


def test_to_seconds():
    assert runez.to_seconds(None) is None
    assert runez.to_seconds("foo") is None
    assert runez.to_seconds("1k") is None
    assert runez.to_seconds("1 m") is None
    assert runez.to_seconds("1 m2s") is None
    assert runez.to_seconds("1m 2") is None
    assert runez.to_seconds("1month") is None
    assert runez.to_seconds([1]) is None  # noqa, verify no crash on bogus type

    assert runez.to_seconds("") == 0
    assert runez.to_seconds(5) == 5
    assert runez.to_seconds("1d1h5s") == 90005
    assert runez.to_seconds("1h  2s") == 3602
    assert runez.to_seconds(" 1h 2s ") == 3602
    assert runez.to_seconds(" 1m ") == 60
    assert runez.to_seconds("1y") == 31556952
    assert datetime.timedelta(seconds=runez.to_seconds("1d1h5s")) == datetime.timedelta(days=1, seconds=3605)
    assert datetime.timedelta(seconds=runez.to_seconds("1w5s")) == datetime.timedelta(days=7, seconds=5)
    assert datetime.timedelta(seconds=runez.to_seconds(" 1w 1s ")) == datetime.timedelta(days=7, seconds=1)

    assert runez.to_seconds(datetime.timedelta(minutes=60)) == 3600
    assert runez.to_seconds(runez.date.UTC.offset) == 0
    tz = runez.timezone_from_text("+0100", default=None)
    assert isinstance(tz, runez.timezone)
    assert runez.to_seconds(tz.offset) == 3600

    with freeze_time("2020-01-02 00:00:12"):
        assert runez.to_seconds("2020-01-01 ") == 86412
        assert runez.to_seconds("2020-01-02 00:00:01") == 11
        assert runez.to_seconds("2020-01-02 00:01:12") == -60



import os
import sys
from unittest.mock import MagicMock, patch

import pytest

import runez
from runez.inspector import auto_import_siblings, AutoInstall, ImportTime


def importable_test_py_files(folder):
    """Finds all .py files in tests/ folder, used for auto-import validation"""
    for fname in os.listdir(folder):
        fpath = os.path.join(folder, fname)
        if os.path.isdir(fpath):
            for x in importable_test_py_files(fpath):
                yield x

        elif fname.endswith(".py"):
            yield fpath


def test_auto_import_siblings():
    # Check that none of these invocations raise an exception
    caller = runez.system.find_caller(depth=1)  # Finds this test as caller
    assert str(caller) == "tests.test_inspector.test_auto_import_siblings"
    with pytest.raises(ImportError):
        # Pretend we're calling auto_import_siblings() from a __main__
        caller.module_name = "__main__"
        assert caller.is_main
        auto_import_siblings(caller=caller)

    with pytest.raises(ImportError):
        # Pretend caller doesn't have a __package__
        caller.package_name = None
        caller.module_name = "foo"
        assert not caller.is_main
        auto_import_siblings(caller=caller)

    py_file_count = len(list(importable_test_py_files(runez.DEV.tests_folder))) - 1  # Remove one to not count tests/__init__.py itself
    imported = auto_import_siblings()
    assert len(imported) == py_file_count

    imported = auto_import_siblings(skip=["tests.secondary"])
    assert len(imported) == py_file_count - 2
    assert "tests.conftest" in imported
    assert "tests.secondary" not in imported
    assert "tests.secondary.test_import" not in imported
    assert "tests.test_system" in imported

    imported = auto_import_siblings(skip=["tests.test_system", "tests.test_serialize"])
    assert len(imported) == py_file_count - 2

    assert "tests.conftest" in imported
    assert "tests.secondary" in imported
    assert "tests.secondary.test_import" in imported
    assert "tests.test_system" not in imported
    assert "tests.test_click" in imported
    assert "tests.test_serialize" not in imported


class SomeClass:
    @AutoInstall("bar")
    def needs_bar(self, msg):
        return "OK: %s" % msg


@AutoInstall("foo")
def needs_foo(msg):
    import foo  # noqa

    return "OK: %s" % msg


def test_auto_install(logged, monkeypatch):
    # Verify that an already present req is a no-op
    AutoInstall("runez").ensure_installed()
    assert not logged

    # Verify failure to install raises abort exception
    with patch("runez.inspector.run", return_value=runez.program.RunResult("failed")):
        with pytest.raises(runez.system.AbortException):
            needs_foo("hello")
        assert "Can't auto-install 'foo': failed" in logged.pop()

    # Verify successful install exercises function call
    with patch("runez.inspector.run", return_value=runez.program.RunResult("OK", code=0)):
        with pytest.raises(ImportError):  # 2nd import attempt raises ImportError (in this case, because we're trying a mocked 'foo')
            needs_foo("hello")
        assert not logged

    # Full successful call
    with patch("runez.inspector.run", return_value=runez.program.RunResult("OK", code=0)):
        assert SomeClass().needs_bar("hello") == "OK: hello"
        assert not logged

    # Mocked successful import
    with patch.dict("sys.modules", foo=MagicMock()):
        with patch("runez.inspector.run", return_value=runez.program.RunResult("OK", code=0)):
            assert needs_foo("hello") == "OK: hello"
            assert not logged

    # Ensure auto-installation is refused unless we have a venv
    monkeypatch.setattr(runez.SYS_INFO, "venv_bin_folder", None)
    with pytest.raises(runez.system.AbortException):
        needs_foo("hello")
    assert "Can't auto-install 'foo' outside of a virtual environment" in logged.pop()


def test_diagnostics_command(cli):
    cli.run("--no-color", "diagnostics")
    assert cli.succeeded
    assert "platform : " in cli.logged
    assert "sys.executable : %s" % runez.short(sys.executable) in cli.logged


@pytest.mark.skipif(sys.version_info[:2] < (3, 7), reason="Available in 3.7+")
def test_importtime():
    """Verify that importing runez remains fast"""
    tos = ImportTime("os")
    tsys = ImportTime("sys")
    trunez = ImportTime("runez")
    assert "runez" in str(trunez)

    assert trunez.cumulative < 3 * tos.cumulative
    assert trunez.cumulative < 3 * tsys.cumulative

    assert trunez.elapsed < 3 * tos.elapsed
    assert trunez.elapsed < 3 * tsys.elapsed


def test_importtime_command(cli):
    cli.run("import-speed")
    assert cli.failed
    assert "Please specify module names, or use --all" in cli.logged

    cli.run("import-speed -i1 --all runez foo_no_such_module runez")
    assert cli.succeeded
    lines = cli.logged.stdout.contents().splitlines()
    assert len([s for s in lines if "runez" in s]) == 1
    assert len([s for s in lines if "foo_no_such_module" in s]) == 1


def test_passthrough(cli):
    cli.run("passthrough")
    assert cli.failed
    assert "Provide command to run" in cli.logged

    cli.run("passthrough", "echo", "hello")
    assert cli.succeeded
    assert "stdout:\nhello" in cli.logged



import pytest

from runez.system import cached_property


class MyObject:

    _global_counter = 0

    @property
    def some_property(self):
        """Example regular (non-cached) property"""
        return "some-value"

    @cached_property
    def foo(self) -> int:
        """Some example property"""
        MyObject._global_counter += 1
        return MyObject._global_counter


def check_props(target):
    # Same outcome regardless whether `target` is an instance (with no cached properties called yet), or a class
    assert sorted(cached_property.properties(target)) == ["foo"]
    assert sorted(cached_property.properties(target, cached_only=False)) == ["foo", "some_property"]
    assert sorted(cached_property.properties(target, cached_only=True)) == ["foo"]


def test_simple_case():
    MyObject._global_counter = None  # Allows to trigger TypeError exception if property is accessed before expected
    cached_property.reset(MyObject)  # No-op when called on a class
    assert isinstance(MyObject.foo, cached_property)
    assert MyObject.foo.__annotations__ == {"return": int}
    assert MyObject.foo.__doc__ == "Some example property"
    assert MyObject.foo.__module__ == "tests.test_cached_property"
    assert MyObject.foo.__name__ == "foo"
    assert cached_property.to_dict(MyObject) is None
    check_props(MyObject)

    obj1 = MyObject()
    check_props(obj1)
    assert cached_property.to_dict(obj1) == {}
    assert "foo" not in obj1.__dict__
    cached_property.reset(obj1)  # no-op, does not trigger computation of property
    assert cached_property.to_dict(obj1, cached_only=False) == {"some_property": "some-value"}

    with pytest.raises(TypeError):
        _ = obj1.foo

    assert "foo" not in obj1.__dict__
    cached_property.reset(obj1)  # no-op
    MyObject._global_counter = 0  # Ensure counter is at known value initially
    assert "foo" not in obj1.__dict__
    assert obj1.foo == 1  # First call
    assert "foo" in obj1.__dict__
    assert cached_property.to_dict(obj1) == {"foo": 1}
    assert cached_property.to_dict(obj1, transform=str) == {"foo": "1"}
    assert cached_property.to_dict(obj1, cached_only=False) == {"foo": 1, "some_property": "some-value"}

    obj2 = MyObject()
    assert cached_property.to_dict(obj2) == {}
    assert cached_property.to_dict(obj2, existing_only=False) == {"foo": 2}  # Freshly computed via existing_only=False
    assert obj2.foo == 2
    assert obj1.foo == 1  # Check that obj1 didn't change
    assert cached_property.to_dict(obj1) == {"foo": 1}

    cached_property.reset(obj1)  # Resets .foo
    assert cached_property.to_dict(obj1) == {}
    assert obj1.foo == 3  # Recomputed after reset
    assert obj2.foo == 2  # Other object remained unaffected
    assert cached_property.to_dict(obj1) == {"foo": 3}

    obj1.foo = 42
    assert obj1.foo == 42  # Setting value works
    assert obj2.foo == 2  # And does not affect other object
    assert cached_property.to_dict(obj1) == {"foo": 42}

    del obj1.foo
    assert "foo" not in obj1.__dict__
    assert cached_property.to_dict(obj1) == {}
    assert obj1.foo == 4  # Recomputed again after delete
    assert "foo" in obj1.__dict__
    assert obj2.foo == 2  # Other object remained unaffected
    assert cached_property.to_dict(obj1) == {"foo": 4}



import pytest

import runez


class Slotted1(runez.Slotted):
    __slots__ = ["a1", "b1"]

    prop1a = runez.AdaptedProperty(default="1a")  # No validation, accepts anything as-is
    prop1b = runez.AdaptedProperty(default="1b", doc="p1b")  # Allows to verify multiple anonymous properties work
    prop1c = runez.AdaptedProperty(caster=int)  # Allows to verify caster approach
    prop1d = runez.AdaptedProperty(type=int, default=123)  # Allows to verify type approach

    @runez.AdaptedProperty(default="p2")
    def prop2(self, value):
        """No validation, accepts anything as-is"""
        return value

    @runez.AdaptedProperty
    def prop3(self, value):
        """Requires something that can be turned into an int"""
        return int(value)

    @runez.AdaptedProperty(default=4)
    def prop4(self, value):
        """Requires something that can be turned into an int"""
        return int(value)

    def _get_defaults(self):
        return runez.UNSET


class Slotted2(runez.Slotted):
    __slots__ = ["name", "other"]


def test_adapted_properties():
    with pytest.raises(AssertionError):
        runez.AdaptedProperty(validator=lambda x: x, caster=int)  # Can't have validator and caster at the same time

    with pytest.raises(AssertionError):
        runez.AdaptedProperty(caster=int, type=int)  # Can't specify both

    s1a = Slotted1()

    # Check class-level properties
    s1ac = s1a.__class__
    assert isinstance(s1ac.prop1a, runez.AdaptedProperty)
    assert s1ac.prop1a.__doc__ is None
    assert s1ac.prop1b.__doc__ == "p1b"
    assert s1ac.prop1c.__doc__ is None
    assert s1ac.prop1d.__doc__ is None
    assert s1ac.prop2.__doc__ == "No validation, accepts anything as-is"
    assert s1ac.prop3.__doc__ == "Requires something that can be turned into an int"

    assert s1a.prop1a == "1a"
    assert s1a.prop1b == "1b"
    assert s1a.prop1c is None
    assert s1a.prop1d == 123
    assert s1a.prop2 == "p2"
    assert s1a.prop3 is None
    assert s1a.prop4 == 4

    # prop1a/b and prop2 have no validators
    s1a.prop2 = 2
    assert s1a.prop2 == 2

    s1a.prop1a = "foo"
    assert s1a.prop1a == "foo"
    assert s1a.prop1b == "1b"  # Verify other anonymous props remain unchanged
    assert s1a.prop1c is None
    assert s1a.prop1d == 123

    s1a.prop1b = 0
    s1a.prop1d = 234
    assert s1a.prop1a == "foo"
    assert s1a.prop1b == 0
    assert s1a.prop1c is None
    assert s1a.prop1d == 234

    s1a.prop1c = "100"  # prop1c has a caster
    assert s1a.prop1a == "foo"
    assert s1a.prop1b == 0
    assert s1a.prop1c == 100

    s1a.prop1c = None  # caster should accept None
    assert s1a.prop1c is None
    with pytest.raises(ValueError):  # but anything other than None must be an int
        s1a.prop1c = "foo"
    assert s1a.prop1c is None

    with pytest.raises(TypeError):
        s1a.prop1d = None  # prop1d uses type=int, and int() does not accept None

    # prop3 and prop4 insist on ints
    s1a.prop3 = "30"
    s1a.prop4 = 40
    assert s1a.prop3 == 30
    assert s1a.prop4 == 40

    with pytest.raises(ValueError):
        s1a.prop3 = "foo"

    # Verify properties stay bound to their object
    s1b = Slotted1(s1a)
    assert s1b.prop1a == "1a"
    assert s1b.prop1b == "1b"
    assert s1b.prop2 == "p2"
    assert s1b.prop3 is None
    assert s1b.prop4 == 4

    s1b.set({"prop1a": "foo"})
    assert s1b.prop1a == "foo"


def test_insights():
    class Foo:
        name = "testing"
        age = 10

    with pytest.raises(ValueError) as exc:
        runez.Slotted.fill_attributes(Foo, {})
    assert "should be instance" in str(exc)

    foo = Foo()
    assert foo.name == "testing"
    assert foo.age == 10
    runez.Slotted.fill_attributes(foo, {"name": "my-name"})
    assert foo.name == "my-name"
    runez.Slotted.fill_attributes(foo, {"name": runez.UNSET})
    assert foo.name == "testing"  # back to class default

    with pytest.raises(AttributeError) as exc:
        runez.Slotted.fill_attributes(foo, {"bar": 5})
    assert "Unknown Foo key 'bar'" in str(exc)


def test_slotted():
    s1a = Slotted1()
    assert s1a.a1 is runez.UNSET

    s1a = Slotted1(a1="a1", b1="b1")
    s1b = Slotted1(s1a)
    assert str(s1b) == "Slotted1(a1=a1, b1=b1)"
    assert s1a.a1 == "a1"
    assert s1a.b1 == "b1"
    assert s1a == s1b

    # Check properties
    assert s1a.prop4 == 4
    assert s1b.prop4 == 4
    s1a.prop4 = 40
    assert s1a.prop4 == 40
    assert s1b.prop4 == 4  # Clone's property did not get modified

    s2 = Slotted2(other=s1a)
    assert s2.name is None
    assert s2.other.a1 == "a1"
    s1a.a1 = "changed-a1"
    assert s2.other.a1 == "a1"

    s2.set(other=s1a)
    assert s2.other is not s1a
    assert s2.other == s1a
    assert s2.other.a1 == "changed-a1"

    s2.set(other="other")
    assert s2.other == "other"

    s2.set(other=s1a)
    assert s2.other == s1a



"""
Convenience logging setup
"""

import faulthandler
import logging
import os
import signal
import sys
import threading
import time
from logging.handlers import RotatingFileHandler, TimedRotatingFileHandler
from typing import List, Optional

from runez.ascii import AsciiAnimation
from runez.convert import to_bytesize, to_int
from runez.date import local_timezone, represented_duration
from runez.file import parent_folder
from runez.system import _R, abort_if, cached_property, decode, DEV, find_caller, flattened, quoted, short, stringified, uncolored
from runez.system import LOG, py_mimic, Slotted, SYS_INFO, ThreadGlobalContext, UNSET


ORIGINAL_CF = logging.currentframe


def formatted(message, *args, **named_values):
    """
    Args:
        message (str): Message to format, support either the '%s' old method, or newer format() method

    Returns:
        (str): Formatted message
    """
    if not named_values:
        if not args:
            return message

        if "%s" in message:
            try:
                return message % args

            except TypeError:
                pass

    try:
        return message.format(*args, **named_values)

    except (IndexError, KeyError):
        return message


class ProgressHandler(logging.Handler):
    """Used to capture logging chatter and show it as progress"""

    level = logging.DEBUG

    @classmethod
    def handle(cls, record):
        """Intercept all log chatter and show it as progress message"""
        LogManager.progress._show_debug(record.getMessage())

    @classmethod
    def emit(cls, record):
        """Not needed"""

    @classmethod
    def createLock(cls):
        """Not needed"""


class ProgressBar:

    def __init__(self, iterable=None, total=None, columns=8, frames=" ▏▎▍▌▋▊▉"):
        self.columns = columns
        self.frames = frames
        self.per_char = 100.0 / columns
        self.blank_char = frames[0]
        self.full_char = frames[-1]
        self.frame_count = len(frames) - 2
        self.per_frame = self.per_char / self.frame_count if self.frame_count > 0 else None
        self.iterable = iterable
        self.parent = None  # type: Optional[ProgressBar] # Parent progress bar, if any
        if total is None and hasattr(iterable, "__len__"):
            total = len(iterable)

        self.total = total
        self.n = None

    def __repr__(self):
        return "%s/%s" % (self.n, self.total)

    def __iter__(self):
        if self.iterable:
            self.start()
            for x in self.iterable:
                yield x
                self.update()

            self.stop()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *_):
        self.stop()

    def start(self):
        """Start tracking progress with this progressbar"""
        if self.n is None:
            self.n = 0
            LogManager.progress._add_progress_bar(self)

    def stop(self):
        """Stop / cleanup this progressbar"""
        if self.n is not None:
            self.n = None
            LogManager.progress._remove_progress_bar(self)

    def update(self, n=1):
        """Manually update the progress bar, advance progress by 'n'"""
        self.n += n

    def rendered(self):
        """Called in spinner thread (lock already acquired)"""
        if self.n is None:
            return None

        if self.total:
            percent = max(int(round(100.0 * self.n / self.total)), 0)
            blanks = 0
            if percent >= 100:
                percent = 100
                bar = self.full_char * self.columns

            else:
                full_chars = int(percent / self.per_char)
                bar = self.full_char * full_chars
                blanks = self.columns - full_chars
                if self.per_frame:
                    blanks -= 1
                    fi = int((percent - full_chars * self.per_char) / self.per_frame)
                    bar += self.frames[fi]

            return "%s%s%s%%" % (bar, self.blank_char * blanks, percent)

    def _remove_parent(self, parent):
        if parent is self.parent:
            self.parent = parent.parent

        elif self.parent is not None:
            self.parent._remove_parent(parent)


class _SpinnerComponent:

    def __init__(self, fps, source, color, adapter=None):
        self.adapter = adapter
        self.source = source  # type: callable
        self.color = color
        self.update_delay = 1.0 / fps
        self.next_update = 0
        self.current_text = None  # type: Optional[str]

    def add_text(self, line, columns):
        """(int): size of text added to 'line' (lock already acquired)"""
        text = self.current_text
        if not text or columns <= 0:
            return 0

        if self.adapter is not None:
            text = self.adapter(text)

        size = len(text)
        if self.adapter is not None or size > columns:
            text = short(text, size=columns)

        text = _R.colored(text, self.color)
        line.append(text)
        return size

    def update_text(self, ts):
        """(int): 1 if changed, 0 otherwise, called by spinner thread (lock already acquired)"""
        if self.next_update < ts:
            self.next_update = ts + self.update_delay
            text = self.source()
            if text is not self.current_text:
                self.current_text = text
                return 1

        return 0


class _SpinnerState:

    def __init__(self, parent, frames, max_columns, message_color, progress_color, spinner_color):
        """
        Args:
            parent (ProgressSpinner): Parent object
            frames (AsciiFrames): Frames to use for spinner
            max_columns (int | None): Optional max number of columns to use
            message_color (str | callable | None): Optional color to use for the message
            progress_color (callable | None): Optional color to use for the spinner
            spinner_color (callable | None): Optional color to use for the spinner
        """
        self.columns = SYS_INFO.terminal.columns - 2
        if max_columns and max_columns > 0:
            self.columns = min(max_columns, self.columns)

        self.frames = _SpinnerComponent(frames.fps, frames.next_frame, spinner_color)
        self.progress_bar = _SpinnerComponent(2, parent._get_progress, progress_color)
        self.message = _SpinnerComponent(2, parent._get_message, message_color, adapter=uncolored)
        self.max_fps = max(2, frames.fps)

    def get_line(self, ts):
        """Called by spinner thread (lock already acquired)"""
        n = self.frames.update_text(ts) + self.progress_bar.update_text(ts) + self.message.update_text(ts)
        if n > 0:
            line = []
            columns = self.columns
            columns -= self.frames.add_text(line, columns)
            columns -= self.progress_bar.add_text(line, columns)
            self.message.add_text(line, columns)
            return " %s" % " ".join(line) if line else ""


class ProgressSpinner:
    """
    Background progress spinner on stderr tty - with an animation conveying that work is being performed.
    This can be enabled with one call:
        runez.log.progress.start()

    A background thread provides spinner animation, will be automatically closed on program exit or when this is called:
        runez.log.progress.stop()

    Optionally, a ProgressBar can be added
    """

    def __init__(self):
        self.is_running = False
        self._current_line = None
        self._fps = 60.0  # Higher fps for _run(), to reduce flickering as much as possible
        self._has_progress_line = False
        self._msg_show = None  # type: Optional[str] # Message coming from show() calls
        self._msg_debug = None  # type: Optional[str] # Message coming from trace() or debug() calls
        self._progress_bar = None  # type: Optional[ProgressBar]
        self._state = None  # type: Optional[_SpinnerState]
        self._stderr_write = None
        self._stdout_write = None
        self._thread = None  # type: Optional[threading.Thread] # Background daemon thread used to display progress

    def show(self, message):
        """
        Args:
            message (str | None): Show 'message' on progress spinner line (this overrides any debug/trace inferred messages)
        """
        with self._lock:
            self._msg_show = message

    def start(self, frames=UNSET, max_columns=140, message_color="dim", progress_color="teal", spinner_color=None):
        """Start a background thread to handle spinner, if stderr is a tty

        Args:
            frames (AsciiFrames | callable | str | None): Frames to use for spinner animation
            max_columns (int | None): Maximum number of terminal columns to use for progress line
            message_color (str | callable | None): Optional color to use for the message part
            progress_color (callable | None): Optional color to use for the progress bar
            spinner_color (callable | None): Optional color to use for the animated spinner
        """
        with self._lock:
            if self._thread is None:
                self._stderr_write = self._original_write(sys.stderr)
                if self._stderr_write is not None:
                    frames = AsciiAnimation.get_frames(frames)
                    self._state = _SpinnerState(self, frames, max_columns, message_color, progress_color, spinner_color)
                    sys.stderr.write = self._on_stderr
                    self._stdout_write = self._original_write(sys.stdout)
                    if self._stdout_write is not None:
                        sys.stdout.write = self._on_stdout

                    self._thread = threading.Thread(target=self._run, name="Progress")
                    self._thread.daemon = True
                    self._thread.start()
                    self.is_running = True
                    LogManager._auto_enable_progress_handler()

    def stop(self):
        """Stop progress spinner thread, called in any thread"""
        with self._lock:
            if self._thread is None:
                return

            self._thread = None

        attempts = 10
        while attempts > 0:
            with self._lock:
                if not self.is_running:  # Wait for thread to exit before cleaning up write functions it may be still using
                    break

            time.sleep(0.05)
            attempts -= 1

        with self._lock:
            self.is_running = False
            LogManager._auto_enable_progress_handler()
            if self._has_progress_line:  # pragma: no cover (hard to cover in tests)
                self._clear_line()
                self._has_progress_line = False

            if sys.stdout.write == self._on_stdout:
                sys.stdout.write = self._stdout_write

            if sys.stderr.write == self._on_stderr:
                sys.stderr.write = self._stderr_write

            self._stderr_write = None
            self._stdout_write = None

    @cached_property
    def _lock(self):
        return threading.RLock()

    def _show_debug(self, message):
        """Show 'message' on next progress line update, called in main thread"""
        with self._lock:
            self._msg_debug = message

    def _get_message(self):
        """Called in spinner thread (lock already acquired)"""
        if self._msg_show is not None:
            return self._msg_show

        return self._msg_debug

    def _get_progress(self):
        """Called in spinner thread (lock already acquired)"""
        if self._progress_bar:
            return self._progress_bar.rendered()

    def _add_progress_bar(self, bar):
        """Called in main thread"""
        with self._lock:
            if bar is not self._progress_bar:
                bar.parent = self._progress_bar
                self._progress_bar = bar

    def _remove_progress_bar(self, bar):
        """Called in main thread"""
        with self._lock:
            if bar is self._progress_bar:
                self._progress_bar = bar.parent

            elif self._progress_bar:
                self._progress_bar._remove_parent(bar)

    @staticmethod
    def _original_write(stream):
        """Called in main thread (lock already acquired)"""
        if SYS_INFO.terminal.isatty(stream):
            return getattr(stream, "write", None)

    def _clean_write(self, write, message):
        """Output 'message' using 'write' function, ensure any pending progress line is cleared first"""
        if message:
            message = decode(message)
            with self._lock:
                if self._has_progress_line:
                    self._clear_line()
                    self._has_progress_line = False

                if self._has_progress_line is False and message.endswith("\n"):
                    self._has_progress_line = None

                write(message)

    def _on_stdout(self, message):
        """Intercepted print() or sys.stdout.write()"""
        self._clean_write(self._stdout_write, message)

    def _on_stderr(self, message):
        """Intercepted sys.stderr.write()"""
        self._clean_write(self._stderr_write, message)

    def _clear_line(self):
        """Called in spinner thread (lock already acquired)"""
        self._write("\r\033[K")

    def _write(self, text):
        """Called in any thread (lock already acquired)"""
        self._stderr_write(text)

    def _run(self):
        """Background thread handling progress reporting and animation"""
        try:
            sleep_delay = 1 / self._fps
            frequency = int(self._fps / self._state.max_fps) - 1
            countdown = 0
            line = None
            while self._thread:
                time.sleep(sleep_delay)
                countdown -= 1
                if countdown < 0 or self._has_progress_line is None:
                    with self._lock:
                        if countdown < 0:
                            countdown = frequency
                            line = self._state.get_line(time.time())

                        if line:
                            self._clear_line()
                            self._write(line)
                            self._write("\r")
                            self._has_progress_line = True

        finally:
            self.is_running = False
            self.stop()


class TraceHandler:
    """
    Allows to optionally provide trace logging, typically activated by an env var, like:
        MY_APP_DEBUG=1 my-app ...
    """

    def __init__(self, prefix, stream):
        self.prefix = prefix
        self.stream = stream

    def trace(self, message):
        """
        Args:
            message (str): Message to trace
        """
        self.stream.write("%s%s%s" % (self.prefix or "", message, "\n" if not message.endswith("\n") else ""))
        self.stream.flush()


class LoggingSnapshot(Slotted):
    """
    Take a snapshot of parts we're modifying in the 'logging' module, in order to be able to restore it as it was
    """

    __slots__ = ["_srcfile", "critical", "fatal", "error", "exception", "warning", "info", "debug"]

    def _seed(self):
        """Seed initial fields"""
        for name in self.__slots__:
            setattr(self, name, getattr(logging, name))

    def restore(self):
        for name in self.__slots__:
            setattr(logging, name, getattr(self, name))


class LogSpec(Slotted):
    """
    Settings to use, you can safely customize these (before calling runez.log.setup), for example:
        runez.log.setup(console_stream=sys.stdout)
    or
        runez.log.spec.set(console_stream=sys.stdout)
        runez.log.setup()
    or
        runez.log.spec.console_stream = sys.stdout
        runez.log.setup()
    """

    # See setup()'s docstring for meaning of each field
    __slots__ = [
        "appname",
        "basename",
        "console_format",
        "console_level",
        "console_stream",
        "context_format",
        "default_logger",
        "dev",
        "project",
        "file_format",
        "file_level",
        "file_location",
        "locations",
        "rotate",
        "rotate_count",
        "timezone",
        "tmp",
    ]

    @property
    def argv(self):
        """str: Command line invocation, represented to show as greeting"""
        return quoted(sys.argv)

    @property
    def pid(self):
        """str: Current process id represented to show as greeting"""
        return os.getpid()

    def usable_location(self):
        """
        Returns:
            str | None: First available usable location
        """
        if not self.should_log_to_file:
            return None

        if self.file_location is not None:
            # Custom location typically provided via --config CLI flag
            return self._auto_complete_filename(self.file_location)

        if self.locations:
            for location in self.locations:
                path = self._auto_complete_filename(location)
                if path:
                    return path

    @property
    def should_log_to_file(self):
        """
        Returns:
            bool: As per the spec, should we be logging to a file?
        """
        if not self.file_format:
            return False

        if self.file_location is not None:
            return bool(self.file_location)

        return bool(self.locations)

    def _props(self, **additional):
        r = dict(argv=self.argv, pid=self.pid)
        r.update(self.to_dict())
        r.update(additional)
        return r

    def _auto_complete_filename(self, location):
        """
        Args:
            location (str | None): Location to auto-complete with {basename}, if it points to a folder

        Returns:
            (str | None): {location}/{basename}
        """
        props = self._props()
        path = _formatted_text(location, props, strict=True)
        if path:
            if os.path.isdir(path):
                filename = _formatted_text(self.basename, props, strict=True)
                if not filename or not is_writable_folder(path):
                    return None

                return os.path.join(path, filename)

            parent = parent_folder(path)
            if not is_writable_folder(parent):
                try:
                    os.mkdir(parent)  # Create only one folder if possible (no mkdir -p)

                except OSError:
                    return None

            if is_writable_folder(parent):
                return path


def is_writable_folder(path):
    return path and os.path.isdir(path) and os.access(path, os.W_OK)


class _ContextFilter(logging.Filter):
    """
    Optional logging filter allowing to inject key/value pairs to every log record.

    In order to activate this:
    - Mention %(context)s in your log format
    - Add key/value pairs via runez.log.context.add_global(), runez.log.context.add_threadlocal()
    """

    def __init__(self, context, name=""):
        """
        Args:
            context (ThreadGlobalContext): Associated context
            name (str): Passed through to parent
        """
        super().__init__(name=name)
        self.context = context

    def filter(self, record):
        """Determines if the record should be logged and injects context info into the record. Always returns True"""
        fmt = LogManager.spec.context_format
        if fmt:
            data = self.context.to_dict()
            if data:
                record.context = fmt % ",".join("%s=%s" % (key, val) for key, val in sorted(data.items()) if key and val)

            else:
                record.context = ""

        return True


def default_log_locations():
    if SYS_INFO.platform_id.is_windows:  # pragma: no cover
        return [os.path.join("{dev}", "log", "{basename}")]

    return ["{dev}/log/{basename}", "/logs/{appname}/{basename}", "/var/log/{basename}"]


class Timeit:
    """Measure how long a decorated function, or context, took took to run"""

    def __init__(self, function=None, color="bold", logger=UNSET, fmt="{function} took {elapsed}"):
        self.__func__ = None
        self.function_name = None
        self.start_time = None
        self.color = color
        self.logger = logger
        self.fmt = fmt
        if callable(function):
            # We're being used as a decorator without args
            self.__func__ = function
            self.function_name = "%s()" % function.__qualname__

        else:
            self.function_name = function

    def __get__(self, instance, owner):
        return _WrappedInstanceFunction(self, instance)

    def __call__(self, *args, **kwargs):
        """
        Args:
            func (callable): We're used as a decorator of a function

        Returns:
            (callable): Decorated function
        """
        if self.__func__:
            with self:
                return self.__func__(*args, **kwargs)

        # We've been used as a decorator with args, and now we're called with the decorated function as argument
        self.__func__ = args[0]
        if not self.function_name:
            self.function_name = "%s()" % self.__func__.__qualname__

        return self

    def __enter__(self):
        self.start_time = time.time()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        msg = self.function_name
        if not msg:
            msg = "%s()" % find_caller()

        logger = _R.rdefault(self.logger, LogManager.spec.default_logger)
        if callable(logger):
            elapsed = time.time() - self.start_time
            elapsed = represented_duration(elapsed)
            msg = _R.colored(msg, self.color)
            elapsed = _R.colored(elapsed, self.color)
            if exc_value:
                description = _R.colored("failed", "red")
                if exc_type is not SystemExit:
                    description += f": {exc_value}"

                msg += f" {description} (ran for {elapsed})"

            else:
                msg = self.fmt.format(function=msg, elapsed=elapsed)

            logger(msg)


class LogManager:
    """
    Global logging context managed by runez.
    There's only one, as multiple contexts would not be useful (logging setup is a global thing)
    """

    # Defaults used to initialize LogSpec instances
    # Use runez.log.override_spec() to change these defaults (do not change directly)
    _default_spec = LogSpec(
        appname=None,
        basename="{appname}.log",
        console_format="%(levelname)s %(message)s",
        console_level=logging.WARNING,
        console_stream=sys.stderr,
        context_format="[[%s]] ",
        default_logger=LOG.debug,
        dev=None,  # Location of development venv where we're currently running from, if any
        project=None,  # Location of source checkout we're currently running from, if any
        file_format="%(asctime)s %(levelname)s %(message)s",
        file_level=logging.DEBUG,
        file_location=None,
        locations=default_log_locations(),
        rotate=None,
        rotate_count=10,
        timezone=local_timezone(),
        tmp=None,
    )

    # Spec defines how logs should be setup()
    # Best way to provide your spec is via: runez.log.setup(), for example:
    #   runez.log.setup(rotate="size:50m")
    spec = LogSpec(_default_spec)

    # Thread-local / global context
    context = ThreadGlobalContext(_ContextFilter)

    # Progress spinner, with animation (on tty only)
    progress = ProgressSpinner()

    # Shown if user tries to run current program as root (and not allowed in setup() call)
    disallow_root_message = "{appname} should not be ran as root!"

    # Below fields should be read-only for outside users, do not modify these
    debug = False
    console_handler: Optional[logging.StreamHandler] = None
    file_handler: Optional[logging.FileHandler] = None  # File we're currently logging to (if any)
    handlers: Optional[List[logging.Handler]] = None
    tracer: Optional[TraceHandler] = None
    used_formats: Optional[str] = None
    faulthandler_signum: Optional[int] = None
    trace_env_var = "TRACE_DEBUG"

    # Convenience decorator/context logging how long a function or section of code took to run
    timeit = Timeit

    _lock = threading.RLock()
    _logging_snapshot = LoggingSnapshot()

    @classmethod
    def set_debug(cls, debug):
        """Useful only as simple callback function, use runez.log.setup() for regular usage"""
        if debug is not UNSET:
            cls.debug = bool(debug)

        return cls.debug

    @classmethod
    def set_dryrun(cls, dryrun):
        """Useful only as simple callback function, use runez.log.setup() for regular usage"""
        _R.set_dryrun(dryrun)
        return _R.is_dryrun()

    @classmethod
    def set_file_location(cls, file_location):
        """Useful only as simple callback function, use runez.log.setup() for regular usage"""
        LogManager.spec.set(file_location=file_location)
        return cls.spec.file_location

    @classmethod
    def setup(
        cls,
        debug=UNSET,
        dryrun=UNSET,
        level=UNSET,
        clean_handlers=UNSET,
        greetings=UNSET,
        appname=UNSET,
        basename=UNSET,
        console_format=UNSET,
        console_level=UNSET,
        console_stream=UNSET,
        context_format=UNSET,
        default_logger=UNSET,
        dev=UNSET,
        file_format=UNSET,
        file_level=UNSET,
        file_location=UNSET,
        locations=UNSET,
        rotate=UNSET,
        rotate_count=UNSET,
        timezone=UNSET,
        tmp=UNSET,
        trace=UNSET,
        allow_root=UNSET,
    ):
        """
        Args:
            debug (bool): Enable debug level logging (overrides other specified levels)
            dryrun (bool): Enable dryrun
            level (int | None): Shortcut to set both `console_level` and `file_level` at once
            clean_handlers (bool): Remove any existing logging.root.handlers
            greetings (str | None): Optional greetings message(s) to log
            appname (str | None): Program's base name, not used directly, just as reference for default 'basename'
            basename (str | None): Base name of target log file, not used directly, just as reference for default 'locations'
            console_format (str | None): Format to use for console log, use None to deactivate
            console_level (int | None): Level to use for console logging
            console_stream (io.TextIOBase | TextIO | None): Stream to use for console log (eg: sys.stderr), use None to deactivate
            context_format (str | None): Format to use for contextual log, use None to deactivate
            default_logger (callable | None): Default logger to use to trace operations such as runez.run() etc
            dev (str | None): Custom folder to use when running from a development venv (auto-determined if None)
            file_format (str | None): Format to use for file log, use None to deactivate
            file_level (int | None): Level to use for file logging
            file_location (str | None): Desired custom file location (overrides {locations} search, handy as a --log cli flag)
            locations (list[str]|None): List of candidate folders for file logging (None: deactivate file logging)
            rotate (str | None): How to rotate log file (None: no rotation, "time:1d" time-based, "size:50m" size-based)
            rotate_count (int): How many rotations to keep
            timezone (str | None): Time zone, use None to deactivate time zone logging
            tmp (str | None): Optional temp folder to use (auto determined)
            trace (str | bool): Env var to enable tracing, example: "DEBUG+| " to trace when $DEBUG defined (+ [optional] "| " as prefix)
            allow_root (bool | None): True allows running as root, None aborts execution if ran as root (default: allowed in docker only)
        """
        with cls._lock:
            cls.set_debug(debug)
            cls.set_dryrun(dryrun)
            cls.spec.set(
                appname=appname,
                basename=basename,
                console_format=console_format,
                console_level=console_level or level,
                console_stream=console_stream,
                context_format=context_format,
                default_logger=default_logger,
                dev=dev,
                file_format=file_format,
                file_level=file_level or level,
                file_location=file_location,
                locations=locations,
                rotate=rotate,
                rotate_count=rotate_count,
                timezone=timezone,
                tmp=tmp,
            )

            cls._auto_fill_defaults()
            if cls.debug:
                cls.spec.console_level = logging.DEBUG
                cls.spec.file_level = logging.DEBUG

            elif level:
                cls.spec.console_level = level
                cls.spec.file_level = level

            root_level = min(flattened(cls.spec.console_level, cls.spec.file_level))
            if root_level and root_level != logging.root.level:
                logging.root.setLevel(root_level)

            if trace is UNSET:
                if cls.tracer is None:
                    trace = cls.trace_env_var

            elif isinstance(trace, str):
                cls.trace_env_var = trace

            if isinstance(trace, str) and "+" in trace:
                p = trace.partition("+")
                cls.enable_trace(p[0], prefix=p[2])

            else:
                cls.enable_trace(trace)

            if cls.handlers is None:
                cls.handlers = []

            cls._setup_console_handler()
            cls._setup_file_handler()
            cls._auto_enable_progress_handler()
            cls._update_used_formats()
            cls._fix_logging_shortcuts()
            if clean_handlers:
                cls.clean_handlers()

            cls.greet(greetings)
            if allow_root is UNSET:
                allow_root = SYS_INFO.is_running_in_docker

            if not allow_root and os.geteuid() == 0:
                message = _formatted_text(cls.disallow_root_message, cls._props(), strict=False)
                if message.endswith("!"):
                    bars = "=" * len(message)
                    message = "\n%s\n%s\n%s\n\n" % (bars, message, bars)

                message = _R.colored(message, "red")
                abort_if(allow_root is None, message)
                LOG.warning(message)

    @staticmethod
    def tests_path(*relative_path):
        """Deprecated, use runez.DEV.tests_path()"""
        return DEV.tests_path(*relative_path)

    @classmethod
    def greet(cls, greetings):
        """
        Args:
            greetings (str | None): Greetings message(s) to log
        """
        if greetings:
            logger = cls.spec.default_logger
            if callable(logger):
                logger(_formatted_text(greetings, cls._props()))

    @classmethod
    def clean_handlers(cls):
        """Remove all non-runez logging handlers"""
        for h in list(logging.root.handlers):
            if h is not cls.console_handler and h is not cls.file_handler and h is not ProgressHandler:
                logging.root.removeHandler(h)

    @classmethod
    def reset(cls):
        """Reset logging as it was before setup(), no need to call this outside of testing, or some very special cases"""
        cls._disable_faulthandler()
        if cls.handlers is not None:
            for handler in cls.handlers:
                logging.root.removeHandler(handler)

            cls.handlers = None

        cls._logging_snapshot.restore()
        cls.context.reset()
        cls.spec = LogSpec(cls._default_spec)
        cls.debug = None
        cls.console_handler = None
        cls.file_handler = None
        cls.progress.stop()
        cls.tracer = None
        cls.used_formats = None

    @classmethod
    def silence(cls, *modules, level=logging.WARNING):
        """
        Args:
            *modules: Modules, or names of modules to silence (by setting their log level to WARNING or above)
            level (int): Logging level to limit logging to for given 'modules'
        """
        for mod in modules:
            name = mod.__name__ if hasattr(mod, "__name__") else mod
            logging.getLogger(name).setLevel(level)

    @classmethod
    def is_using_format(cls, markers, used_formats=None):
        """
        Args:
            markers (str): Space separated list of markers to look for
            used_formats (str): Formats to consider (default: cls.used_formats)

        Returns:
            (bool): True if any one of the 'markers' is seen in 'used_formats'
        """
        if used_formats is None:
            used_formats = cls.used_formats

        if not markers or not used_formats:
            return False

        return any(marker in used_formats for marker in flattened(markers, split=" "))

    @classmethod
    def enable_faulthandler(cls, signum=getattr(signal, "SIGUSR1", None)):
        """Enable dumping thread stack traces when specified signals are received, similar to java's handling of SIGQUIT

        Note: this must be called from the surviving process in case of daemonization.

        Args:
            signum (int | None): Signal number to register for full thread stack dump (use None to disable)
        """
        with cls._lock:
            if not signum:
                cls._disable_faulthandler()
                return

            if not cls.file_handler:
                return

            cls.faulthandler_signum = signum
            dump_file = cls.file_handler.stream
            faulthandler.enable(file=dump_file, all_threads=True)  # noqa
            faulthandler.register(signum, file=dump_file, all_threads=True, chain=False)  # noqa

    @classmethod
    def override_spec(cls, **settings):
        """Override 'spec' and '_default_spec' with given values"""
        cls._default_spec.set(**settings)
        cls.spec.set(**settings)

    @classmethod
    def enable_trace(cls, spec, prefix=":: ", stream=UNSET):
        """
        Args:
            spec (str | bool | None): If string given, enable tracing when corresponding env var is set to a non-empty value
            prefix (str | None): Prefix to use for trace messages (default: ":: ")
            stream: Where to trace (by default: current 'console_stream' if configured, otherwise sys.stderr)
        """
        prior = cls.tracer
        if spec is not UNSET:
            if spec and (not isinstance(spec, str) or spec in os.environ):
                cls.tracer = TraceHandler(prefix, stream or cls.spec.console_stream or sys.stderr)

            else:
                cls.tracer = None

        return prior

    @classmethod
    def resolved_dryrun(cls, dryrun):
        """
        Args:
            dryrun (bool | runez.Undefined | None): Optionally overridden current dryrun setting

        Returns:
            (bool): Resolved value for dryrun
        """
        return _R.resolved_dryrun(dryrun)

    @classmethod
    def trace(cls, message, *args):
        """
        Args:
            message (str): Message to trace
        """
        if cls.tracer or cls.progress.is_running:
            message = formatted(message, *args)
            cls.progress._show_debug(message)
            if cls.tracer:
                cls.tracer.trace(message)

    @classmethod
    def hdry(cls, message, dryrun=UNSET, logger=UNSET):
        """Handle dryrun, allows to handle dryrun=UNSET with a code pattern of the form:

            if runez.log.hdry("it was a dryrun"):
                return

        Args:
            message (str | callable | None): Message to log
            dryrun (bool | UNSET | None): Optionally override current dryrun setting
            logger (callable | bool | None): Logger to use, True to print(), False to trace(), None to disable log chatter

        Returns:
            (bool): True if we were indeed in dryrun mode, and we logged the message
        """
        return _R.hdry(dryrun, logger, message)

    @classmethod
    def _props(cls):
        if cls.file_handler and cls.file_handler.baseFilename:
            location = cls.file_handler.baseFilename

        elif not cls.spec.should_log_to_file:
            location = "file log disabled"

        elif cls.spec.file_location:
            location = "given location '{file_location}' is not usable"

        else:
            location = "no usable locations from {locations}"

        return cls.spec._props(location=location)

    @classmethod
    def _auto_enable_progress_handler(cls):
        if cls.progress.is_running:
            if ProgressHandler not in logging.root.handlers:
                logging.root.handlers.append(ProgressHandler)  # noqa

        elif ProgressHandler in logging.root.handlers:
            logging.root.handlers.remove(ProgressHandler)  # noqa

    @classmethod
    def _update_used_formats(cls):
        cls.used_formats = None
        for handler in (cls.console_handler, cls.file_handler):
            fmt = _get_fmt(handler)
            if fmt:
                cls.used_formats = "%s %s" % (cls.used_formats or "", fmt)
                cls.used_formats = cls.used_formats.strip()

    @classmethod
    def _setup_console_handler(cls):
        fmt = _canonical_format(cls.spec.console_format)
        level = cls.spec.console_level
        target = cls.spec.console_stream
        existing = cls.console_handler
        if existing is None or _get_fmt(existing) != fmt or existing.level != level or existing.stream != target:
            if existing is not None:
                cls.handlers.remove(existing)
                logging.root.removeHandler(existing)

            if target:
                cls.console_handler = cls._add_handler(logging.StreamHandler(target), fmt, level)

    @classmethod
    def _setup_file_handler(cls):
        fmt = _canonical_format(cls.spec.file_format)
        level = cls.spec.file_level
        target = cls.spec.usable_location()
        existing = cls.file_handler
        if existing is None or _get_fmt(existing) != fmt or existing.level != level or existing.baseFilename != target:
            if existing is not None:
                cls.handlers.remove(existing)
                logging.root.removeHandler(existing)

            if target:
                cls.file_handler = cls._add_handler(_get_file_handler(target, cls.spec.rotate, cls.spec.rotate_count), fmt, level)

    @classmethod
    def _add_handler(cls, new_handler, fmt, level):
        if fmt:
            new_handler.setFormatter(logging.Formatter(fmt))

        if level:
            new_handler.setLevel(level)

        logging.root.addHandler(new_handler)
        cls.handlers.append(new_handler)
        return new_handler

    @classmethod
    def _auto_fill_defaults(cls):
        """Late auto-filled missing defaults (caller's value kept if provided)"""
        if not cls.spec.appname:
            cls.spec.appname = SYS_INFO.program_name

        if not cls.spec.dev:
            cls.spec.dev = DEV.venv_folder

        if not cls.spec.project:
            cls.spec.project = DEV.project_folder

    @classmethod
    def _disable_faulthandler(cls):
        if cls.faulthandler_signum:
            faulthandler.unregister(cls.faulthandler_signum)
            faulthandler.disable()
            cls.faulthandler_signum = None

    @classmethod
    def _fix_logging_shortcuts(cls):
        """
        Fix standard logging shortcuts to correctly report logging module.

        This is only useful if you:
        - actually use %(name) and care about it being correct
        - you would still like to use the logging.info() etc shortcuts

        So basically you'd like to write this:
            import logging
            logging.info("hello")

        Instead of this:
            import logging
            LOG = logging.getLogger(__name__)
            LOG.info("hello")
        """
        if cls.is_using_format("%(context)"):
            cls.context.enable(True)
            for handler in cls.handlers:
                handler.addFilter(cls.context.filter)

        else:
            for handler in cls.handlers:
                handler.removeFilter(cls.context.filter)

            cls.context.enable(False)

        if cls.is_using_format("%(pathname) %(filename) %(funcName) %(module)"):
            logging._srcfile = cls._logging_snapshot._srcfile

        else:
            logging._srcfile = None

        logging.logProcesses = cls.is_using_format("%(process)")
        logging.logThreads = cls.is_using_format("%(thread) %(threadName)")

        getframe = getattr(sys, "_getframe", None)
        if not isinstance(logging.info, _LogWrap) and getframe is not None:
            logging.critical = _LogWrap(logging.CRITICAL)
            logging.fatal = logging.critical
            logging.error = _LogWrap(logging.ERROR)
            logging.exception = _LogWrap(logging.ERROR, exc_info=True)
            logging.warning = _LogWrap(logging.WARNING)
            logging.info = _LogWrap(logging.INFO)
            logging.debug = _LogWrap(logging.DEBUG)
            logging.log = _LogWrap.log


class _LogWrap:
    """Allows to correctly report caller file/function/line from convenience calls such as logging.info()"""

    def __init__(self, level, exc_info=None):
        self.level = level
        self.exc_info = exc_info
        py_mimic(self, getattr(logging, logging.getLevelName(level).lower()))

    @staticmethod
    def log(level, msg, *args, **kwargs):
        getframe = getattr(sys, "_getframe", None)
        offset = kwargs.pop("_stack_offset", 1)
        name = getframe(offset).f_globals.get("__name__")
        logger = logging.getLogger(name)
        try:
            logging.currentframe = lambda: getframe(3 + offset)
            logger.log(level, msg, *args, **kwargs)

        finally:
            logging.currentframe = ORIGINAL_CF

    def __call__(self, msg, *args, **kwargs):
        kwargs.setdefault("exc_info", self.exc_info)
        kwargs.setdefault("_stack_offset", 2)
        self.log(self.level, msg, *args, **kwargs)


def _replace_and_pad(fmt, marker, replacement):
    """
    Args:
        fmt (str): Format to tweak
        marker (str): Marker to replace
        replacement (str): What to replace marker with

    Returns:
        (str): Resulting format, with marker replaced
    """
    if marker not in fmt:
        return fmt

    if replacement:
        return fmt.replace(marker, replacement)

    # Remove mention of 'marker' in 'fmt', including leading space (if present)
    fmt = fmt.replace("%s " % marker, marker)
    return fmt.replace(marker, "")


def _canonical_format(fmt):
    """
    Args:
        fmt (str | None): Format specification

    Returns:
        (str | None): Canonical version of format
    """
    if not fmt:
        return fmt

    return _replace_and_pad(fmt, "%(timezone)s", LogManager.spec.timezone)


def _format_recursive(key, value, definitions, max_depth):
    m = _R.lc.rx_format_markers.search(value)
    if not m:
        return value

    if max_depth > 1 and value and "{" in value:
        try:
            value = value.format(**definitions)
            return _format_recursive(key, value, definitions, max_depth=max_depth - 1)

        except KeyError:
            pass

    return value


def _formatted_text(text, props, strict=False, max_depth=3):
    """
    Args:
        text (str): Text with '{...}' placeholders to be resolved
        props (dict): Available values

    Returns:
        (str): '{...}' placeholders resolved from given `props`
    """
    if not text:
        return text

    if text.startswith("~"):
        text = os.path.expanduser(text)

    if "{" not in text:
        return text

    definitions = {}
    markers = _R.lc.rx_format_markers.findall(text)
    while markers:
        key = markers.pop()
        if key in definitions:
            continue

        val = props.get(key)
        if strict and val is None:
            return None

        val = stringified(val) if val is not None else "{%s}" % key
        markers.extend(m for m in _R.lc.rx_format_markers.findall(val) if m not in definitions)
        definitions[key] = val

    if not max_depth or not isinstance(max_depth, int) or max_depth <= 0:
        return text

    result = dict((k, _format_recursive(k, v, definitions, max_depth)) for k, v in definitions.items())
    return text.format(**result)


def _get_fmt(handler):
    return handler and handler.formatter and handler.formatter._fmt


def _get_file_handler(location, rotate, rotate_count):
    """
    Args:
        location (str | None): Log file path
        rotate (str | None): How to rotate, examples:
            time:midnight - Rotate at midnight
            time:15s - Rotate every 15 seconds
            time:2h - Rotate every 2 hours
            time:7d - Rotate every 7 days
            size:20m - Rotate every 20MB
            size:1g - Rotate every 1MB
        rotate_count (int): How many backups to keep

    Returns:
        (logging.FileHandler): Associated handler
    """
    if not rotate:
        return logging.FileHandler(location)

    kind, _, mode = rotate.partition(":")

    if not mode:
        raise ValueError("Invalid 'rotate' (missing kind): %s" % rotate)

    if kind == "time":
        if mode == "midnight":
            return TimedRotatingFileHandler(location, when="midnight", backupCount=rotate_count)

        timed = "shd"
        if mode[-1].lower() not in timed:
            raise ValueError("Invalid 'rotate' (unknown time spec): %s" % rotate)

        interval = to_int(mode[:-1])
        if interval is None:
            raise ValueError("Invalid 'rotate' (time range not an int): %s" % rotate)

        return TimedRotatingFileHandler(location, when=mode[-1], interval=interval, backupCount=rotate_count)

    if kind == "size":
        size = to_bytesize(mode)
        if size is None:
            raise ValueError("Invalid 'rotate' (size not a bytesize): %s" % rotate)

        return RotatingFileHandler(location, maxBytes=size, backupCount=rotate_count)

    raise ValueError("Invalid 'rotate' (unknown type): %s" % rotate)


class _WrappedInstanceFunction:

    def __init__(self, function, instance):
        self.__func__ = function
        self.instance = instance

    def __call__(self, *args, **kwargs):
        self.__func__(self.instance, *args, **kwargs)



import os
import re
import sys

import pytest

import runez
import runez.conftest


def sample_main():
    args = sys.argv[1:]
    runez.log.trace("Running main() with: %s" % args)
    if args:
        args = runez.flattened(args, shellify=True)
        if args[0] == "TypeError":
            # Raise a TypeError
            len(42)

        exit_code = runez.to_int(args[0])
        if exit_code is not None:
            # When first arg is a number, call sys.exit() with that number
            if len(args) > 1:
                print(" ".join(args[1:]))

            sys.exit(exit_code)

        if args[0] == "Exception":
            # Raise a generic exception
            raise Exception("crashed: %s" % args[1:])

        if args[0] == "AssertionError":
            assert False, "oops, something went wrong"

        if args[0] == "exit":
            # exit without explicit code
            sys.exit(" ".join(args[1:]))

        if args[0] == "quiet":
            # Don't output anything
            sys.exit(0)

    # Simulate some output
    return "%s %s" % (os.path.basename(sys.argv[0]), " ".join(args))


def test_cli_uninitialized(cli, monkeypatch):
    from runez.conftest import cli as cli_fixture

    monkeypatch.setattr(cli_fixture, "default_main", None)
    with pytest.raises(AssertionError):
        # No main provided
        cli.run("hello no main")


def test_crash(cli):
    cli.main = sample_main
    cli.run(["Exception", "hello with main"])
    assert cli.failed
    assert cli.match("crashed...hello")
    assert cli.match("Exited with stacktrace:")

    with pytest.raises(AssertionError):
        cli.run(["AssertionError"])

    cli.run("TypeError")
    assert cli.failed
    assert cli.match("TypeError: ... has no len")

    cli.run("exit", "some message")
    assert cli.failed
    cli.expect_messages("some message", "!stacktrace")

    cli.expect_failure("Exception hello", "crashed...hello", "Exited with stacktrace:", "!this message shouldn't appear")

    cli.run(["successful hello"])
    assert cli.succeeded
    assert cli.match("successful hello")

    cli.run(["Exception", "hello again"])
    assert cli.failed
    assert not cli.match("hello with main")
    assert not cli.match("successful hello")
    assert cli.match("hello again")

    cli.run(1, "hello")
    assert cli.failed
    assert "hello" in cli.logged.stdout
    assert "Exited with stacktrace" not in cli.logged

    cli.run(2)
    assert cli.failed
    assert cli.exit_code == 2
    assert not cli.logged

    with pytest.raises(AssertionError):
        # No captures specified
        assert cli.match("crashed...hello", stdout=False, stderr=False)

    with pytest.raises(AssertionError):
        # Expect success failed
        cli.expect_success("Exception", "hello")

    with pytest.raises(AssertionError):
        # Unexpected message seen in output
        cli.expect_failure(["Exception", "hello"], "!crashed...hello")

    with pytest.raises(AssertionError):
        # Expected message not seen in output
        cli.expect_failure(["Exception", "hello"], "this message shouldn't appear")


def test_edge_cases(temp_folder, monkeypatch):
    # Exercise dev folder determination code
    info = runez.system.DevInfo()
    info.tests_folder = "./bar/baz"
    info.venv_folder = "./foo"
    assert info.project_path() is None

    runez.touch("setup.py", logger=None)
    del info.project_folder
    p = info.project_path()
    assert p == "."


def test_invalid_main(cli):
    cli.run("hello", main="-mno_such_module")
    assert cli.failed
    assert "No module named no_such_module" in cli.logged.stderr

    with pytest.raises(AssertionError):
        cli.run("hello", main="no_such_script.foo")

    with pytest.raises(AssertionError):
        cli.run("foo", main=["invalid"])


def test_script_invocations(cli):
    # This unfortunately does is not enough to detect code coverage (due to cwd being a temp folder I think)
    cli.run("--help", main="-mrunez")
    assert cli.succeeded
    assert "usage: python -mrunez [-h]" in cli.logged
    assert "Set of sample commands" in cli.logged

    cli.run("--help", main="src/runez/__main__.py")
    assert cli.succeeded
    assert "/runez [-h]" in cli.logged
    assert "Set of sample commands" in cli.logged

    # Below will properly make test coverage detect properly that we did execute code in __main__.py
    cli.exercise_main("-mrunez", "src/runez/__main__.py")
    cli.exercise_main("extra-validations")  # Checks also that script is found even in tests/ subfolder

    with pytest.raises(AssertionError):
        cli.exercise_main("failed-help")


def test_success(cli, monkeypatch):
    cli.main = sample_main

    # Verify that project folder works properly
    tests = os.path.dirname(__file__)
    project_folder = os.path.abspath(os.path.join(tests, ".."))

    assert cli.project_folder == project_folder
    assert runez.DEV.project_folder == project_folder
    assert cli.tests_folder == tests
    assert cli.tests_path("foo.txt") == os.path.join(tests, "foo.txt")
    assert cli.project_path() == project_folder
    assert cli.project_path("foo") == os.path.join(project_folder, "foo")

    cli.run("quiet")
    assert cli.succeeded
    assert not cli.logged
    assert cli.match(".*", regex=True) is None

    cli.run("quiet", trace=True)
    assert cli.succeeded
    assert ":: Running main() with: ['quiet']" in cli.logged

    cli.run("--dryrun hello", exe="bar/foo")
    assert cli.succeeded
    assert cli.logged.stdout.contents() == "foo --dryrun hello"
    assert not cli.logged.stderr

    cli.run("--dryrun hello")
    assert cli.succeeded
    assert cli.logged.stdout.contents() == "pytest --dryrun hello"
    assert not cli.logged.stderr
    assert cli.match("el+", regex=True)
    assert not cli.match("EL+", regex=True)
    assert cli.match("EL+", regex=re.IGNORECASE)

    cli.expect_success("hello world", "hello", "el+", regex=True)
    m = cli.match("hello ...l")
    assert str(m) == "hello worl"
    assert cli.match("el+", regex=True).match == "ell"
    assert cli.match(re.compile("hel+o")).match == "hello"
    assert cli.match("h...")
    assert cli.match("h...", regex=True)
    assert not cli.match("h...", regex=False)
    assert not cli.match("Hello")
    assert cli.match("Hello", regex=re.IGNORECASE)

    cli.run([""])
    assert cli.succeeded
    assert cli.logged.stdout.contents().strip() == os.path.basename(sys.argv[0])
    assert not cli.logged.stderr
    assert not cli.match("hello")



import sys
from collections import namedtuple
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

import runez
from runez.http import ForbiddenHttpError, GlobalHttpCalls, MockResponse, RestClient, RestResponse, urljoin


EXAMPLE = RestClient("https://example.com")
CacheState = namedtuple("CacheState", "cached hits misses updates")


class MockBackend:
    def __init__(self):
        self.cache = {}
        self.cache_hits = 0
        self.cache_misses = 0
        self.cache_updates = 0
        self.last_expire = None

    @property
    def state(self):
        return CacheState(len(self.cache), self.cache_hits, self.cache_misses, self.cache_updates)

    def delete(self, cache_key):
        if cache_key in self.cache:
            self.cache_updates += 1
            del self.cache[cache_key]

    def get(self, cache_key):
        if cache_key in self.cache:
            self.cache_hits += 1
            return self.cache[cache_key]

        self.cache_misses += 1

    def set(self, cache_key, data, expire=None):
        self.cache_updates += 1
        self.last_expire = expire
        self.cache[cache_key] = data


@EXAMPLE.mock({
    "test/README.txt?a=b": "Hello",
})
def test_cache(monkeypatch):
    c = RestClient.std_diskcache()
    assert c is None

    with patch("runez.http.DEV.current_test", return_value=False):
        monkeypatch.setitem(sys.modules, "diskcache", MagicMock())
        c = RestClient.std_diskcache()
        assert isinstance(c.cache_backend, MagicMock)
        assert c.default_expire == 3600

        cm = MockBackend()
        c.cache_backend = cm
        assert cm.state == CacheState(cached=0, hits=0, misses=0, updates=0)
        client = EXAMPLE.sub_client("")
        client.cache_wrapper = c
        response = client.get_response("foo")
        assert not response.ok
        assert cm.state == CacheState(cached=0, hits=0, misses=1, updates=0)

        response = client.get_response("test/README.txt", params={"a": "b"})
        assert response.text == "Hello"
        assert cm.cache["https://example.com/test/README.txt?a=b"] == response
        assert cm.last_expire == 3600
        assert cm.state == CacheState(cached=1, hits=0, misses=2, updates=1)

        response2 = client.get_response("test/README.txt", params={"a": "b"})
        assert response2 is response
        assert cm.state == CacheState(cached=1, hits=1, misses=2, updates=1)

        client.purge("test/README.txt", params={"a": "b"})
        assert cm.cache_updates == 2
        assert cm.state == CacheState(cached=0, hits=1, misses=2, updates=2)

        response3 = client.get_response("test/README.txt", params={"a": "b"}, expire=5)
        assert response3 is not response
        assert response3.text == response.text
        assert cm.last_expire == 5
        assert cm.state == CacheState(cached=1, hits=1, misses=3, updates=3)


@GlobalHttpCalls.allowed
def test_decorator_allowed():
    assert GlobalHttpCalls.is_forbidden() is False
    with GlobalHttpCalls(allowed=False) as mg:
        assert GlobalHttpCalls.is_forbidden() is True
        assert str(mg) == "forbidden"

    assert GlobalHttpCalls.is_forbidden() is False


def test_default_disabled():
    assert GlobalHttpCalls.is_forbidden() is True
    with pytest.raises(ForbiddenHttpError) as exc:
        client = RestClient()
        client.head("https://example.com")
    assert "https://example.com" in str(exc)


@GlobalHttpCalls.forbidden
def test_decorator_forbidden():
    assert GlobalHttpCalls.is_forbidden() is True
    with GlobalHttpCalls(allowed=True) as mg:
        assert GlobalHttpCalls.is_forbidden() is False
        assert str(mg) == "allowed"

    assert GlobalHttpCalls.is_forbidden() is True


@EXAMPLE.mock({
    "test/README.txt": "Hello",
})
def test_download(temp_folder, logged):
    assert str(EXAMPLE) == "https://example.com"
    client = EXAMPLE.sub_client("test/")
    assert str(client) == "https://example.com/test/"
    r = client.put("README.txt", data=Path("foo"), fatal=False, dryrun=True)
    assert r.ok
    assert r.json() == {"message": "dryrun PUT https://example.com/test/README.txt"}
    assert "Would PUT" in logged.pop()

    with pytest.raises(FileNotFoundError):
        # fatal=False addresses http(s) communications only, not existence of files that are referred to by caller
        client.put("README.txt", data=Path("foo"), fatal=False, dryrun=False)
    assert not logged

    assert client.download("test.zip", "test.zip", dryrun=True).ok
    assert "Would download" in logged.pop()

    assert client.download("foo/test.zip", "test.zip", fatal=False).status_code == 404
    assert "404" in logged.pop()

    assert client.download("README.txt", "README.txt", fatal=False).ok
    assert "GET https://example.com/test/README.txt [200]" in logged.pop()
    assert list(runez.readlines("README.txt")) == ["Hello"]

    # With checksum validation
    assert client.download("README.txt#sha256=a123", "README.txt", fatal=False) is None
    assert "Deleted README.txt" in logged
    assert "sha256 differs for README.txt: expecting a123, got " in logged.pop()

    r = client.download("README.txt#sha256=185f8db32271fe25f561a6fc938b2e264306ec304eda518007d1764826381969", "README.txt", fatal=False)
    assert r.ok

    assert client.download("README.txt#md5=8b1a9953c4611296a827abf8c47804d7", "README.txt", fatal=False).ok
    assert client.download("README.txt#sha1=f7ff9e8b7bb2e09b70935a5d785e0cc5d9d0abf0", "README.txt", fatal=False).ok
    c = "3615f80c9d293ed7402687f94b22d58e529b8cc7916f8fac7fddf7fbd5af4cf777d3d795a7a00a16bf7e7f3fb9561ee9baae480da9fe7a18769e71886b03f315"
    assert client.download(f"README.txt#sha512={c}", "README.txt", fatal=False).ok

    # Not considered checksum, url fragment left as-is
    assert client.download("README.txt#sha2=bar", "README.txt", fatal=False).status_code == 404

    client.decompress("foo/test.tar.gz#sha256=123", "my-folder", dryrun=True)
    assert "Would untar test.tar.gz -> my-folder" in logged.pop()


def test_edge_cases():
    assert urljoin("", "") == ""
    assert urljoin("a", "") == "a"
    assert urljoin("", "b") == "b"
    assert urljoin("a", "b") == "a/b"
    assert urljoin("a#fragment", "b") == "a#fragment/b"
    assert urljoin("http://example.net/a", "b") == "http://example.net/a/b"
    assert urljoin("http://example.net/a", "https://example.com/b") == "https://example.com/b"
    assert urljoin("http://example.net/a/#/b", "https://example.com/b") == "https://example.com/b"
    assert urljoin("http://example.net/a/#/b", "c") == 'http://example.net/a/#/b/c'
    assert urljoin("http://example.net/a#b", "c") == 'http://example.net/a#b/c'


@EXAMPLE.mock({})
@RestClient.handler.mock
@EXAMPLE.mock
@RestClient.handler.mock({})
def test_files(temp_folder):
    """yolo"""
    # Exercise download code path
    sample = Path("README.txt")
    r = EXAMPLE.download("README", sample, fatal=False)
    assert r.status_code == 404
    assert r.url == "https://example.com/README"

    with EXAMPLE.mock({"README": "hello"}) as mm:
        assert str(mm) == "tests.test_http.test_files started, 1 specs"
        assert str(mm.stack) == "RequestsHandler mock active, 1 specs [depth: 5]"
        EXAMPLE.download("README", sample)
        assert list(runez.readlines(sample)) == ["hello"]

        # Use local README.txt, which should get opened/closed appropriately
        # Exercise data=Path(...) code path, headers are temporarily used
        r = EXAMPLE.post("README", headers={"foo": "bar"}, data=sample)
        assert isinstance(r, RestResponse)
        assert r.ok

        # Exercise filepaths= code path
        r = EXAMPLE.post("README", filepaths={"sample": sample})
        assert isinstance(r, RestResponse)
        assert r.ok


@EXAMPLE.mock
@RestClient.handler.mock("https://example.com/tt", {"test": 205})
def test_handler_mock(logged):
    session = RestClient("https://example.com", handler=EXAMPLE.handler)
    assert session.head("tt/test", fatal=False).status_code == 205
    assert session.head("foo", fatal=False).status_code == 404

    with EXAMPLE.mock({"foo": 201, "tt/test": 206}) as mm:
        mock_stack = mm.stack
        assert str(mm.stack) == "RequestsHandler mock active, 2 specs [depth: 3]"
        assert str(mm) == "tests.test_http.test_handler_mock started, 2 specs"
        assert session.head("foo", fatal=False).status_code == 201
        assert session.head("tt/test", fatal=False).status_code == 206  # Comes from closest mock context

    assert str(mm) == "tests.test_http.test_handler_mock stopped, 2 specs"
    assert mm.stack is None
    assert str(mock_stack) == "RequestsHandler mock active, 1 specs [depth: 2]"
    assert session.head("foo", fatal=False).status_code == 404
    assert session.head("tt/test", fatal=False).status_code == 205  # Reverts back to prev mock


def test_reporting():
    # Verify reasonable extraction of error messages
    assert RestResponse.extract_message(None) is None
    assert RestResponse.extract_message(" foo ") == "foo"
    assert RestResponse.extract_message({"message": " oops "}) == "oops"
    assert RestResponse.extract_message({"error": " oops "}) == "oops"
    assert RestResponse.extract_message({"errors": " oops "}) == "oops"
    assert RestResponse.extract_message({"errors": [{"error": " nested "}]}) == "nested"


def dynamic_call(method, url):
    if url.endswith("-a"):
        return 201, "invalid json"  # Simulate request not return valid json

    return ["bar", method]  # Implied status 200


@EXAMPLE.mock({
    "foo-bar": {"foo": "bar"},  # status 200 implied, payload is a dict
    "bad-request": (400, dict(error="oops", msg="more info")),  # status 400, with sample error
    "server-crashed": (500, "failed"),  # status 500, with optional content as well
    "dynamic-a": dynamic_call,  # status and payload will come from function call
    "dynamic-b": dynamic_call,
    "explicit": MockResponse(202, "explicit RestResponse"),
    "fail1": Exception,
    "fail2": Exception("oops"),
})
def test_rest():
    with runez.CaptureOutput(trace=True) as logged:
        session = RestClient("https://example.com", headers={"test": "testing"})
        assert len(session.headers) == 2
        assert session.headers["test"] == "testing"
        assert session.headers["User-Agent"]

        assert session.url_exists("foo-bar") is True
        assert session.delete("foo-bar").ok
        assert "DELETE https://example.com/foo-bar [200]" in logged.pop()

        assert session.purge("foo-bar").ok
        assert "PURGE https://example.com/foo-bar [200]" in logged.pop()

        assert session.get("foo-bar") == {"foo": "bar"}
        assert session.get("https://example.com/foo-bar") == {"foo": "bar"}
        assert "GET https://example.com/foo-bar [200]" in logged.pop()
        assert session.post("foo-bar").ok
        assert "POST https://example.com/foo-bar [200]" in logged.pop()
        assert session.put("foo-bar").ok
        assert "PUT https://example.com/foo-bar [200]" in logged.pop()

        assert not session.post("bad-request", fatal=False).ok
        assert "POST https://example.com/bad-request [400] oops" in logged.pop()

        # Status 500 in mock spec does NOT impact dryrun
        assert session.post("server-crashed", dryrun=True).ok
        assert "Would POST" in logged.pop()

        # But does impact actual (no dryrun) run
        with pytest.raises(runez.system.AbortException):
            session.get("server-crashed", fatal=True)
        assert "GET https://example.com/server-crashed [500]" in logged.pop()
        r = session.get_response("server-crashed", fatal=False, logger=None)
        assert not r.ok
        assert r.status_code == 500

        assert session.url_exists("") is False
        assert session.url_exists("not-found") is False
        assert session.get("not-found", fatal=False, logger=None) is None
        assert session.head("not-found", fatal=False, logger=None).status_code == 404
        logged.pop()

        assert str(session.get_response("dynamic-a", logger=None)) == "<Response [201]>"
        r = session.get_response("dynamic-b", logger=None)
        assert r.description(size=12) == "GET https://..."
        assert not logged

        r = session.put("explicit")
        assert r.method == "PUT"
        assert str(r) == "<Response [202]>"

        with pytest.raises(Exception) as exc:
            session.get("fail1")
        assert "Simulated crash" in str(exc)

        with pytest.raises(Exception) as exc:
            session.get("fail2")
        assert "oops" in str(exc)



import random
import threading

from runez.thread import thread_local_property, ThreadLocalSingleton


class MySingleton(ThreadLocalSingleton):
    def __init__(self):
        self.some_number = random.randint(0, 100000) + random.randint(0, 100000)


def new_singleton():
    return MySingleton()


def test_singleton():
    s1 = MySingleton()
    s2 = MySingleton()
    assert s1 is s2
    assert s1 is new_singleton()

    s3 = run_in_new_thread(MySingleton)
    s4 = run_in_new_thread(new_singleton)
    assert type(s1) is type(s3)
    assert type(s1) is type(s4)
    assert s1 is not s3
    assert s1 is not s4
    assert s3 is not s4


class SomeClass:

    times_called = 0

    @thread_local_property
    def thread_id(self):
        """Testing thread local properties"""
        self.times_called += 1
        return random.randint(0, 100000) + random.randint(0, 100000)

    def get_thread_id(self):
        return self.thread_id


class WrappedRun:
    def __init__(self, target, args, kwargs):
        self.target = target
        self.args = args
        self.kwargs = kwargs
        self.result = None

    def run(self):
        self.result = self.target(*self.args, **self.kwargs)


def run_in_new_thread(target, *args, **kwargs):
    wrapper = WrappedRun(target, args, kwargs)
    t = threading.Thread(target=wrapper.run)
    t.start()
    t.join()
    return wrapper.result


def test_thread_local_property():
    prop = SomeClass.thread_id
    assert prop.__name__ == "thread_id"
    assert prop.__doc__ == "Testing thread local properties"

    obj = SomeClass()
    main_tid = obj.get_thread_id()
    assert obj.times_called == 1
    assert main_tid
    assert main_tid == obj.get_thread_id()
    assert obj.times_called == 1

    tid1 = run_in_new_thread(obj.get_thread_id)
    assert tid1 != main_tid
    assert obj.times_called == 2

    tid2 = run_in_new_thread(obj.get_thread_id)
    assert tid2 != main_tid
    assert tid2 != tid1
    assert obj.times_called == 3



import datetime
import logging

import runez
from runez.schema import Any, Boolean, Date, Datetime, Dict, Enum, Float, Integer, List, String, Struct, UniqueIdentifier
from runez.serialize import Serializable, SerializableDescendants, with_behavior


def test_any():
    any = Any()
    assert any.problem(None) is None
    assert any.problem("a") is None
    assert any.problem(4) is None
    assert any.problem([1, 2]) is None
    assert any.problem(object()) is None


def test_boolean():
    b = Boolean()
    assert b.problem(None) is None
    assert b.problem("a") is None
    assert b.problem(4) is None
    assert b.problem([1, 2]) is None
    assert b.problem(object()) is None

    assert b.converted(None) is None

    assert b.converted(True) is True
    assert b.converted(False) is False

    assert b.converted("a") is False
    assert b.converted("False") is False
    assert b.converted("true") is True
    assert b.converted("Y") is True
    assert b.converted("yes") is True

    assert b.converted(0) is False
    assert b.converted(4) is True
    assert b.converted(123.4) is True
    assert b.converted("123.4") is True
    assert b.converted("123.4a") is False

    assert b.converted([]) is False
    assert b.converted({}) is False
    assert b.converted([1, 2]) is True
    assert b.converted({"a": "b"}) is True

    assert b.converted(object()) is True


def test_date():
    dd = Date()
    assert str(dd) == "Date"
    assert dd.problem(None) is None
    assert dd.problem({}) == "expecting date, got '{}'"
    assert dd.converted(0) == datetime.date(1970, 1, 1)
    assert dd.converted("2019-09-02") == datetime.date(2019, 9, 2)
    assert dd.converted("2019-09-02 01:02:03") == datetime.date(2019, 9, 2)

    dd = Datetime()
    assert str(dd) == "Datetime"
    assert dd.problem(None) is None
    assert dd.problem({}) == "expecting datetime, got '{}'"
    assert dd.converted(0) == datetime.datetime(1970, 1, 1, tzinfo=runez.date.UTC)
    assert dd.converted("2019-09-02") == datetime.datetime(2019, 9, 2, tzinfo=runez.date.UTC)
    assert dd.converted("2019-09-02 01:02:03") == datetime.datetime(2019, 9, 2, 1, 2, 3, tzinfo=runez.date.UTC)


def test_dict():
    dd = Dict()
    assert str(dd) == "Dict[Any, Any]"
    assert dd.problem({}) is None
    assert dd.problem({1: "a"}) is None
    assert dd.problem({"a": 1}) is None

    dd = Dict(String)
    assert str(dd) == "Dict[String, Any]"
    assert dd.problem({}) is None
    assert dd.problem({1: "a"}) == "key: expecting string, got '1'"
    assert dd.problem({"a": 1}) is None

    dd = Dict(String, Integer)
    assert str(dd) == "Dict[String, Integer]"
    assert dd.problem(5) == "expecting dict, got '5'"
    assert dd.problem({}) is None
    assert dd.problem({1: "a"}) == "key: expecting string, got '1'"
    assert dd.problem({"a": "b"}) == "value: expecting int, got 'b'"
    assert dd.problem({"a": 1}) is None

    assert dd.converted({"a": 1}) == {"a": 1}
    assert dd.converted({"a": "1"}) == {"a": 1}

    dd = Dict(String, List(Integer))
    assert str(dd) == "Dict[String, List[Integer]]"
    assert dd.problem({}) is None
    assert dd.problem({"a": "b"}) == "value: expecting list, got 'b'"
    assert dd.problem({"a": ["1"]}) is None
    assert dd.problem({"a": ["b"]}) == "value: expecting int, got 'b'"


def test_enum():
    ee = Enum("foo bar")
    assert str(ee) == "Enum[bar, foo]"
    assert ee.problem(None) is None
    assert ee.problem("foo") is None
    assert ee.problem("x") == "'x' is not one of Enum[bar, foo]"
    assert ee.problem(1) == "'1' is not one of Enum[bar, foo]"


def test_list():
    ll = List()
    assert str(ll) == "List[Any]"
    assert ll.problem([]) is None
    assert ll.problem(["a"]) is None
    assert ll.problem([1, 2]) is None
    assert ll.problem({1, "2"}) is None

    assert ll.converted([1, "2"]) == [1, "2"]

    ll = List(Integer)
    assert str(ll) == "List[Integer]"
    assert ll.problem([]) is None
    assert ll.problem(["a"]) == "expecting int, got 'a'"
    assert ll.problem([1, 2]) is None
    assert ll.problem((1, 2)) is None
    assert ll.problem({1, "2"}) is None

    assert ll.converted([1, "2"]) == [1, 2]
    assert ll.converted((1, 2)) == [1, 2]
    assert sorted(ll.converted({1, "2"})) == [1, 2]


def test_number():
    ff = Float()
    assert str(ff) == "Float"
    assert ff.problem(None) is None
    assert ff.problem(5) is None
    assert ff.problem(5.3) is None
    assert ff.problem("foo") == "expecting float, got 'foo'"
    assert ff.problem([]) == "expecting float, got '[]'"

    assert ff.converted("5.4") == 5.4
    assert ff.converted(5) == 5.0
    assert ff.converted("0o10") == 8.0


class Car(Serializable, with_behavior(extras=(logging.info, "foo bar"))):
    make = String
    serial = UniqueIdentifier
    year = Integer

    def __init__(self, make=None, serial=None, year=None):
        self.make = make
        self.serial = serial
        self.year = year


class Hat(Struct):
    size = Integer(default=1)


class SpecializedCar(Car):
    """Used to test that ._meta.behavior is passed on properly to descendants"""

    hats = List(Hat)  # Test list of serializables


class Person(Serializable, with_behavior(strict=logging.error, hook=logging.info)):
    age = Date
    fingerprint = UniqueIdentifier(Integer)
    name = String(default="joe")

    car = Car
    hat = Hat

    def __init__(self, name=None):
        self.name = name


class GPerson(Person):
    """Used to test that ._meta.behavior is passed on properly to descendants"""

    age = Integer  # Change type on purpose
    group = Integer


def test_serializable(logged):
    assert len(list(SerializableDescendants.children(Car))) == 2
    assert len(list(SerializableDescendants.children(Person))) == 2
    assert len(list(SerializableDescendants.children(SpecializedCar))) == 1

    assert str(Serializable._meta.behavior) == "lenient"

    assert str(Car._meta.behavior) == "extras: function 'info', ignored extras: [foo, bar]"
    assert str(SpecializedCar._meta.behavior) == "extras: function 'debug'"  # extras are NOT inherited
    assert Car._meta.attributes_by_type(String) == ["make", "serial"]
    assert Car._meta.attributes_by_type(Integer) == ["year"]
    assert SpecializedCar._meta.attributes_by_type(Integer) == ["year"]
    assert SpecializedCar._meta.attributes_by_type(List) == ["hats"]

    assert str(Person._meta) == "Person (5 attributes, 0 properties)"
    assert str(Person._meta.behavior) == "strict: function 'error', extras: function 'debug', hook: function 'info'"
    # `hook` is inherited
    assert str(GPerson._meta.behavior) == "strict: function 'error', extras: function 'debug', hook: function 'info'"

    # Verify that most specific type wins (GPerson -> age)
    assert Person._meta.attributes_by_type(Integer) == ["fingerprint"]
    assert Person._meta.attributes_by_type(Date) == ["age"]
    assert GPerson._meta.attributes_by_type(Integer) == ["age", "fingerprint", "group"]
    assert GPerson._meta.attributes_by_type(Date) is None

    car = Car(year=2020)
    assert car.to_dict() == {"year": 2020}

    car = Car.from_dict({"foo": 1, "baz": 2, "serial": "bar"})
    assert car.serial == "bar"
    assert "foo" not in logged
    assert "Extra content given for Car: baz" in logged.pop()
    assert car.to_dict() == {"serial": "bar"}

    pp = Person()
    assert pp.age is None
    assert pp.fingerprint is None
    assert pp.name is None  # overridden by Person.__init__()
    assert pp.car is None
    assert pp.to_dict() == {}

    pp = Person.from_dict({"car": "foo", "fingerprint": "foo"})
    assert pp.fingerprint == "foo"  # Bogus value still used because 'strict' setting does not raise an exception
    assert "Can't deserialize Person.car: expecting compliant dict, got 'foo'" in logged
    assert "Can't deserialize Person.fingerprint: expecting int, got 'foo'" in logged.pop()

    Person.from_dict({"hat": {"size": "foo"}})
    assert "Can't deserialize Person.hat: expecting int, got 'foo'" in logged.pop()

    pp = Person.from_dict({"age": "2019-01-01", "car": {"make": "Honda", "year": 2010}, "fingerprint": 5})
    assert pp.age.year == 2019
    assert pp.car.make == "Honda"
    assert pp.car.year == 2010
    assert pp.fingerprint == 5
    assert pp.hat is None
    assert pp.to_dict() == {"age": "2019-01-01", "car": {"make": "Honda", "year": 2010}, "fingerprint": 5, "name": "joe"}

    # Default `name` from schema is respected, because we didn't call Person.__init__() explicitly
    pp = Person.from_dict({"age": 1567296012, "hat": {"size": "5"}})
    assert pp.age.year == 2019
    assert pp.hat.size == 5
    assert pp.to_dict() == {"age": "2019-09-01", "name": "joe", "hat": {"size": 5}}

    Person.from_dict({"car": {"make": "Honda", "foo": "bar"}})
    assert "'foo' is not an attribute of Car" in logged.pop()

    Person.from_dict({"age": "foo"})
    assert "Can't deserialize Person.age: expecting date, got 'foo'" in logged.pop()

    pp = Person.from_dict({"car": {"make": "Honda", "foo": "bar"}})
    assert pp.car.make == "Honda"
    assert "'foo' is not an attribute of Car" in logged.pop()


def test_string():
    ss = String()
    assert str(ss) == "String"
    assert ss.problem(None) is None
    assert ss.problem("foo") is None
    assert ss.problem(1) == "expecting string, got '1'"

    assert ss.converted(None) is None
    assert ss.converted("foo") == "foo"
    assert ss.converted(1) == "1"
    assert ss.converted([1, 2]) == "[1, 2]"
